import fred.src.utils as utils
import fred.src.web_interface.yaml_to_wi_object as yto
import fred.src.web_interface.factors_and_conditions as fac_cond
import fred.src.web_interface.wi_utils as wi_utils
import copy

disabled_fields = []

# TODO: code refactoring + documentation


def edit_wi_object(path, key_yaml, read_in_whitelists):
    """
    This function fills an empty wi object with the information of a metadata
    file
    :param key_yaml:
    :param path: the path containing the metadata file
    :return: wi_object: the filled wi object
    """
    # TODO: as Parameter at get_info
    meta_yaml = utils.read_in_yaml(path)
    whitelist_object = {}

    if meta_yaml is not None:
        if "path" in meta_yaml:
            meta_yaml.pop("path")
        empty_object = yto.get_empty_wi_object(key_yaml, read_in_whitelists)
        wi_object = {}
        wi_object["all_factors"], real_val = get_all_factors(meta_yaml, key_yaml)
        wi_object["publication_records"] = get_publication_records(meta_yaml)
        for part in empty_object:
            if part not in ["all_factors", "publication_records"]:
                wi_object[part], whitelist_object = new_fill(
                    meta_yaml[part],
                    empty_object[part],
                    key_yaml[part],
                    whitelist_object,
                    real_val,
                    read_in_whitelists,
                )

    else:
        wi_object = yto.get_empty_wi_object(key_yaml, read_in_whitelists)

    wi_object["whitelists"] = whitelist_object
    wi_object = get_filenames(wi_object, meta_yaml)
    return wi_object


def get_publication_records(meta_yaml):
    publications = list(utils.find_keys(meta_yaml, "publication"))
    publication_records = {}
    for elem in publications:
        for pub in elem:
            if pub["pubmed_id"] not in publication_records:
                publication_records[pub["pubmed_id"]] = pub
    return publication_records


def get_filenames(wi_object, meta_yaml):
    old_filenames = {}
    old_sample_names = {}
    settings = list(utils.find_keys(meta_yaml, "experimental_setting"))
    if len(settings) > 0:
        for setting in settings:
            for stg in setting:
                if "setting_id" in stg:
                    setting_id = stg["setting_id"]
                    bio_reps = list(utils.find_keys(stg, "samples"))
                    if len(bio_reps) > 0:
                        for elem in bio_reps:
                            for sample in elem:
                                if "sample_name" in sample:
                                    key = f'{setting_id}_{sample["sample_name"]}'
                                    tech_rep = list(
                                        utils.find_keys(sample, "technical_replicates")
                                    )
                                    if len(tech_rep) > 0:
                                        if "filenames" in tech_rep[0]:
                                            old_filenames[key] = tech_rep[0][
                                                "filenames"
                                            ]
                                        if "sample_name" in tech_rep[0]:
                                            old_sample_names[key] = tech_rep[0][
                                                "sample_name"
                                            ]
    wi_object["old_sample_names"] = old_sample_names
    wi_object["old_filenames"] = old_filenames
    return wi_object


def new_fill(
    meta_yaml, wi_object, key_yaml, whitelist_object, real_val, read_in_whitelists
):

    if isinstance(meta_yaml, dict):

        if "headers" in wi_object:
            fill_key = "value"
            filled_value = ""
            for header in wi_object["headers"].split(" "):
                filled_value = filled_value + " " + meta_yaml[header]
            filled_value = filled_value.lstrip(" ").rstrip(" ")
        else:
            if wi_object["position"].split(":")[-1] == "experimental_setting":
                fill_key = "input_fields"
                filled_value, whitelist_object = fill_experimental_setting(
                    wi_object,
                    meta_yaml,
                    key_yaml["value"],
                    whitelist_object,
                    real_val,
                    read_in_whitelists,
                )
            else:
                fill_key = "input_fields"
                filled_value = copy.deepcopy(wi_object["input_fields"])
                for i in range(len(filled_value)):
                    if filled_value[i]["position"].split(":")[-1] in meta_yaml:
                        filled_value[i], whitelist_object = new_fill(
                            meta_yaml[filled_value[i]["position"].split(":")[-1]],
                            filled_value[i],
                            key_yaml["value"][
                                filled_value[i]["position"].split(":")[-1]
                            ],
                            whitelist_object,
                            real_val,
                            read_in_whitelists,
                        )

    elif isinstance(meta_yaml, list):
        fill_key = "list_value"
        filled_value = []
        for i in range(len(meta_yaml)):
            f_val, whitelist_object = new_fill(
                meta_yaml[i],
                copy.deepcopy(wi_object),
                key_yaml,
                whitelist_object,
                real_val,
                read_in_whitelists,
            )

            # TODO: WTF?
            if "input_fields" in f_val:
                f_val = f_val["input_fields"]
                if (
                    "position" in f_val
                    and f_val["position"].split(":")[-1] == "experimental_setting"
                ):
                    f_val["input_disabled"] = True
            else:
                f_val = f_val["value"]
            filled_value.append(f_val)

    else:
        fill_key = "value"
        if wi_object["input_type"] == "date":
            filled_value = wi_utils.str_to_date(meta_yaml)
        else:
            filled_value = meta_yaml

    if "input_type" in wi_object and wi_object["input_type"] == "single_autofill":
        fill_key = "list_value"

    wi_object[fill_key] = filled_value

    if "special_case" in key_yaml and "edit" in key_yaml["special_case"]:
        if key_yaml["special_case"]["edit"] == "not editable":
            wi_object["input_disabled"] = True
            wi_object["delete_disabled"] = True
        elif key_yaml["special_case"]["edit"] == "not removable":
            wi_object["delete_disabled"] = True
            if key_yaml["list"]:
                wi_object["fixed_length"] = (
                    len(filled_value) if filled_value is not None else None
                )
                if "input_type" in wi_object and wi_object["input_type"] in [
                    "select",
                    "group_select",
                ]:
                    wi_object["non_editable_val"] = filled_value
    return wi_object, whitelist_object


def fill_experimental_setting(
    wi_object, meta_yaml, key_yaml, whitelist_object, real_val, read_in_whitelists
):
    organism = ""
    filled_object = []
    for j in range(len(wi_object["input_fields"])):
        f = copy.deepcopy(wi_object["input_fields"][j])
        for key in meta_yaml:

            if wi_object["input_fields"][j]["position"].split(":")[-1] == key:
                if key == "experimental_factors":
                    pass
                elif key == "conditions":
                    sample_keys = list(utils.find_keys(key_yaml, "samples"))
                    if len(sample_keys) > 0:

                        sample, whitelists = yto.parse_empty(
                            sample_keys[0],
                            "experimental_setting:conditions:biological_"
                            "replicates:samples",
                            key_yaml,
                            {"organism": organism},
                            read_in_whitelists,
                            get_whitelist_object=True,
                        )
                        sample = sample["input_fields"]

                        conditions = []

                        for cond in meta_yaml[key]:
                            samples = []
                            split_cond = utils.split_cond(cond["condition_name"])
                            sample_name = utils.get_short_name(
                                cond["condition_name"],
                                {},
                                key_yaml,
                                read_in_whitelists=read_in_whitelists,
                            )
                            input_fields = fac_cond.get_samples(
                                split_cond,
                                copy.deepcopy(sample),
                                real_val,
                                key_yaml,
                                sample_name,
                                organism,
                                read_in_whitelists,
                            )
                            if "samples" in cond["biological_replicates"]:
                                for s in cond["biological_replicates"]["samples"]:
                                    filled_keys = []
                                    for k in s:
                                        if s[k] is not None:
                                            # TODO: dict (disease usw.)
                                            # TODO: real_val
                                            if isinstance(s[k], list):
                                                for elem in s[k]:
                                                    if (
                                                        k,
                                                        elem,
                                                    ) not in filled_keys and (
                                                        k,
                                                        elem,
                                                    ) not in split_cond:
                                                        filled_keys.append((k, elem))
                                            elif (k, s[k]) not in split_cond:
                                                filled_keys.append((k, s[k]))
                                    sample_index = int(
                                        s["sample_name"].split("_")[-1].replace("b", "")
                                    )
                                    cond_sample_name = (
                                        f"{sample_name}_" f"{sample_index}"
                                    )
                                    filled_sample = copy.deepcopy(input_fields)
                                    filled_sample = fac_cond.get_samples(
                                        filled_keys,
                                        filled_sample,
                                        real_val,
                                        key_yaml,
                                        cond_sample_name,
                                        organism,
                                        read_in_whitelists,
                                        is_factor=False,
                                    )
                                    samples.append(copy.deepcopy(filled_sample))
                            title, readd = fac_cond.get_condition_title(split_cond)
                            d = {
                                "correct_value": cond["condition_name"],
                                "title": title,
                                "readd": readd,
                                "input_disabled": False,
                                "position": "experimental_setting:condition",
                                "list": True,
                                "mandatory": True,
                                "list_value": samples,
                                "desc": "",
                                "input_fields": input_fields,
                            }
                            if (
                                "special_case" in sample_keys[0]
                                and "edit" in sample_keys[0]["special_case"]
                            ):
                                if (
                                    sample_keys[0]["special_case"]["edit"]
                                    == "not editable"
                                ):
                                    d["input_disabled"] = True
                                    d["delete_disabled"] = True
                                elif (
                                    sample_keys[0]["special_case"]["edit"]
                                    == "not removable"
                                ):
                                    d["delete_disabled"] = True
                                    if sample_keys[0]["list"]:
                                        d["fixed_length"] = (
                                            len(samples)
                                            if samples is not None
                                            else None
                                        )
                            conditions.append(d)
                        # TODO: not editable and not removable for conditions?
                        f["list_value"] = conditions
                        whitelist_object[organism] = whitelists

                else:

                    if "headers" in f and isinstance(meta_yaml[key], dict):
                        new_val = parse_headers(meta_yaml[key], f["headers"])
                    else:
                        new_val = meta_yaml[key]

                    if key == "organism":
                        organism = new_val

                    if "whitelist_keys" in f:
                        new_val = parse_whitelist_keys(
                            meta_yaml[key],
                            f["whitelist_keys"],
                            utils.get_whitelist(
                                key,
                                {"organism": organism},
                                whitelist_object=read_in_whitelists,
                            ),
                        )

                    if "list" in f and f["list"]:
                        f["list_value"] = new_val
                    else:
                        f["value"] = new_val
                filled_object.append(f)
    return filled_object, whitelist_object


def parse_headers(value, headers):

    if isinstance(headers, dict):
        header = None
        for k in headers:
            if sorted(headers[k].split(" ")) == sorted(list(value.keys())):
                header = headers[k].split(" ")
                break
    else:
        header = headers.split(" ")

    if header is not None:
        val = ""
        for h in header:
            val = f'{val}{" " if val != "" else ""}{value[h]}'
        value = val

    return value


def parse_whitelist_keys(value, whitelist_keys, whitelist):

    for key in whitelist_keys:
        if f"{value} ({key})" in whitelist:
            value = f"{value} ({key})"
            break

    return value


# TODO: value unit?
def get_all_factors(meta_yaml, real_val):
    """
    This function creates an object containing all experimental factors from a
    metadata yaml to be stored in a wi object
    :param real_val:
    :param meta_yaml: the read in metadata file
    :return: all_factors: an object containing all experimental factors
    """
    all_factors = []
    for setting in meta_yaml["experimental_setting"]:
        setting_factors = []
        for factors in setting["experimental_factors"]:
            setting_fac = {"factor": factors["factor"]}
            w = utils.get_whitelist(factors["factor"], setting)
            setting_fac["values"] = []
            for elem in factors["values"]:
                value = elem
                if w and "headers" in w:
                    if "headers" not in setting_fac:
                        setting_fac["headers"] = w["headers"]
                    if isinstance(elem, dict):
                        value = parse_headers(value, w["headers"])

                if w and "whitelist_keys" in w:
                    if "whitelist_keys" not in setting_fac:
                        setting_fac["whitelist_keys"] = w["whitelist_keys"]
                    if "whitelist_type" in w and w["whitelist_type"] == "plain_group":
                        value = parse_whitelist_keys(
                            value, w["whitelist_keys"], w["whitelist"]
                        )

                if (
                    isinstance(elem, dict)
                    and len(list(elem.keys())) == 2
                    and "unit" in elem
                    and "value" in elem
                ):
                    value = f'{elem["value"]}{elem["unit"]}'

                if value != elem:
                    if isinstance(elem, dict):
                        # rewrite the value into a string
                        val = "|".join([f'{key}:"{elem[key]}"' for key in elem])
                        val = f'{factors["factor"]}:{"{"}{val}{"}"}'
                        real_val[val] = value
                    else:
                        real_val[elem] = value
                # TODO: nach oben rücken + rekursiv (eigene Funktion)
                elif isinstance(factors["values"], dict):
                    value = {value: factors["values"][value]}
                setting_fac["values"].append(value)
            setting_factors.append(setting_fac)
        all_factors.append(setting_factors)
    return all_factors, real_val



import plotly.graph_objects as go
import fred.src.utils as utils
from fred.src.heatmap.dash_utils import *
import plotly.express as px
import os
import base64
import io
from PIL import Image
import numpy


def get_heatmap(
    path,
    keys_yaml,
    show_setting_id=True,
    only_factors=False,
    mode="samples",
    labels="factors",
    background=False,
    sample_labels=False,
    condition_labels=False,
    transpose=False,
    drop_defaults=False,
):
    (
        settings,
        experimental_factors,
        organisms,
        max_vals,
        options_pretty,
        annotated_dict,
        max_annotation,
        conditions,
    ) = get_data(
        path, keys_yaml, mode=mode, drop_defaults=drop_defaults, transpose=transpose
    )

    techniques = list(utils.find_keys(path, "techniques"))
    setting_techniques = {}
    if len(techniques) > 0:
        techniques = techniques[0]
        for elem in techniques:
            try:
                sett_id = elem["setting"]
                try:
                    technique = elem["technique"]
                    setting_techniques[sett_id] = technique
                except KeyError:
                    pass
            except KeyError:
                pass

    heatmaps = []

    for value in settings:
        fig = None
        if len(settings[value]) > 0:

            options = [
                key.replace("_num", "")
                for key in settings[value]
                if key.endswith("_num")
            ]

            if only_factors:
                sorter = experimental_factors[value]
            else:
                sorter = experimental_factors[value] + [
                    o for o in options if o not in experimental_factors[value]
                ]
            df_empty = settings[value].dropna(axis=1, how="all")
            sorter = [x for x in sorter if x in df_empty.columns]
            val_sorter = [0]
            for elem in sorter:
                if elem in max_annotation[value]:
                    val_sorter.append(
                        max(
                            val_sorter[-1] + (50 if not transpose else 150),
                            val_sorter[-1] + (15 * max_annotation[value][elem]) + 20,
                        )
                    )
                else:
                    val_sorter.append(val_sorter[-1] + (50 if not transpose else 150))

            df = [settings[value][f"{key}_num"] for key in sorter]
            annotated = [
                annotated_dict[value][key]
                for key in sorter
                if key in annotated_dict[value]
            ]
            if transpose:
                annotated = numpy.transpose(annotated)
            option_text = []
            label_text = []
            for key in sorter:
                if labels == "all" or (
                    labels == "factors"
                    and key in annotated_dict[value]
                    and key in experimental_factors[value]
                ):
                    label_vals = []
                    for elem in annotated_dict[value][key]:
                        splitted = elem.split("<br>")
                        new_val = []
                        for val in splitted:
                            if len(val) > 20:
                                new_val.append(f"{val[:18]}...")
                            else:
                                new_val.append(val)
                        label_vals.append("<br>".join(new_val))
                    label_text.append(label_vals)
                else:
                    label_text.append([""] * len(annotated_dict[value][key]))

            for option in sorter:
                if option in experimental_factors[value]:
                    option_text.append(color("red", f"<b>{options_pretty[option]}</b>"))
                else:
                    option_text.append(color("black", options_pretty[option]))

            plotly_colors = []

            colors_scales = [
                px.colors.qualitative.G10,
                px.colors.qualitative.Plotly,
                px.colors.qualitative.D3,
                px.colors.qualitative.T10,
            ]

            i = 0
            while len(plotly_colors) < max_vals[value]:
                plotly_colors += [
                    x
                    for x in colors_scales[i % len(colors_scales)]
                    if x not in ["#DC3912", "#EF553B", "#D62728", "#F58518"]
                ]
                i += 1

            colors = [[0, "white"]]
            for i in range(1, max_vals[value] + 1):
                colors.append([i * 1 / (max_vals[value]), plotly_colors[i - 1]])
                if i * 1 / (max_vals[value]) != 1:
                    colors.append([((i * 1) / (max_vals[value])) + 0.001, "white"])

            if transpose:
                if mode == "samples":
                    if condition_labels:
                        if sample_labels:

                            y_value = [
                                settings[value]["condition_label"],
                                settings[value]["sample_label"],
                            ]
                        else:
                            y_value = [
                                settings[value]["condition_label"],
                                settings[value]["sample_index"],
                            ]
                    else:
                        if sample_labels:
                            y_value = [
                                settings[value]["condition_index"],
                                settings[value]["sample_label"],
                            ]
                        else:
                            y_value = [
                                settings[value]["condition_index"],
                                settings[value]["sample_index"],
                            ]
                else:
                    if condition_labels:
                        y_value = settings[value]["condition_label"]
                    else:
                        y_value = settings[value]["condition_index"]

                x_value = val_sorter
                label_text = numpy.transpose(label_text)
                df = numpy.transpose(df)
            else:
                if mode == "samples":
                    if condition_labels:
                        if sample_labels:
                            x_value = [
                                settings[value]["condition_label"],
                                settings[value]["sample_label"],
                            ]
                        else:
                            x_value = [
                                settings[value]["condition_label"],
                                settings[value]["sample_index"],
                            ]
                    else:
                        if sample_labels:
                            x_value = [
                                settings[value]["condition_index"],
                                settings[value]["sample_label"],
                            ]
                        else:
                            x_value = [
                                settings[value]["condition_index"],
                                settings[value]["sample_index"],
                            ]
                else:
                    if condition_labels:
                        x_value = settings[value]["condition_label"]
                    else:
                        x_value = settings[value]["condition_index"]
                y_value = val_sorter

            heatmap = [
                go.Heatmap(
                    z=df,
                    zmin=0,
                    zmax=max_vals[value],
                    x=x_value,
                    y=y_value,
                    showscale=False,
                    customdata=annotated,
                    text=label_text,
                    texttemplate="%{text}",
                    hovertemplate="%{customdata}",
                    hoverongaps=False,
                    colorscale=colors,
                ),
            ]

            data_input = heatmap

            if mode == "samples":
                if sample_labels:
                    sample_label_height = max(settings[value]["sample_label_height"])
                else:
                    sample_label_height = 1
            else:
                sample_label_height = 0

            if condition_labels:
                condition_label_height = max(settings[value]["condition_label_height"])
            else:
                condition_label_height = 1

            my_cell_width = 150
            # my_cell_height = max(50, (15*max_annotation[value])+20)
            if transpose:
                top_margin = 100
                bottom_margin = 20
                left_margin = (150 if condition_labels else 100) + (
                    150 if sample_labels else 100
                )
                right_margin = 20
                my_width = val_sorter[-1]
                my_height = 50 * (
                    len(settings[value]["sample_index"])
                    if mode == "samples"
                    else len(settings[value]["condition_index"])
                )
            else:
                top_margin = (
                    15 * sample_label_height + 15 * condition_label_height + 70
                )  # minimum 100
                bottom_margin = 20
                left_margin = 200
                right_margin = 20
                my_height = val_sorter[-1]  # my_cell_height*len(sorter)

                if mode == "samples":
                    my_width = my_cell_width * len(settings[value]["sample_index"])
                else:
                    my_width = my_cell_width * len(settings[value]["condition_index"])
            organism_path = os.path.join(
                os.path.dirname(__file__),
                "..",
                "..",
                "images",
                f"{organisms[value]}.png",
            )
            images = None

            if os.path.isfile(organism_path):
                plotly_logo = base64.b64encode(open(organism_path, "rb").read())
                imgdata = base64.b64decode(plotly_logo)
                im = Image.open(io.BytesIO(imgdata))
                im_width, im_height = im.size
                y_side = 100
                my_ysize = y_side / my_height
                x_side = y_side * im_width / im_height
                my_xsize = x_side / my_width

                left_margin = max(left_margin, x_side)

                images = [
                    dict(
                        source="data:image/png;base64,{}".format(plotly_logo.decode()),
                        xref="paper",
                        yref="paper",
                        x=0,
                        y=1,
                        sizey=my_ysize,
                        sizex=my_xsize,
                        xanchor="right",
                        yanchor="bottom",
                    )
                ]

            between = []
            technique_list = (
                setting_techniques[value] if value in setting_techniques else None
            )
            if technique_list is not None:
                technique_list = f' - {", ".join(technique_list)}'
            else:
                technique_list = ""
            for i in range(len(val_sorter) - 1):
                between.append((val_sorter[i + 1] + val_sorter[i]) / 2)
            layout = go.Layout(
                images=images,
                height=my_height + top_margin + bottom_margin,
                width=my_width + left_margin + right_margin,
                margin=dict(
                    l=left_margin, r=right_margin, t=top_margin, b=bottom_margin
                ),
                autosize=False,
                title=dict(
                    text=(
                        f"<b>Setting {value}{technique_list}</b>"
                        if show_setting_id
                        else ""
                    ),
                    font=dict(size=15),
                    automargin=False,
                    yref="container",
                    x=(left_margin + (0.1 * 150))
                    / (my_width + left_margin + right_margin),
                    y=1
                    - (30 / (my_height + top_margin + bottom_margin) * 100)
                    / (my_height + top_margin + bottom_margin),
                    xanchor="left",
                    yanchor="top",
                    subtitle=dict(
                        text=f"Organism: {organisms[value]}",
                        font=dict(size=14, lineposition="under"),
                    ),
                ),
                plot_bgcolor="rgb(255, 255, 255)" if background else "rgba(0,0,0,0)",
                paper_bgcolor="rgb(255, 255, 255)" if background else "rgba(0,0,0,0)",
                xaxis=(
                    dict(
                        side="top",
                        automargin=False,
                    )
                    if not transpose
                    else dict(
                        tickmode="array",
                        ticktext=option_text,
                        tickvals=between,
                        side="top",
                        automargin=False,
                    )
                ),
                yaxis=(
                    dict(
                        tickmode="array",
                        ticktext=option_text,
                        tickvals=between,
                        autorange="reversed",
                        automargin=False,
                    )
                    if not transpose
                    else dict(
                        automargin=False,
                        autorange="reversed",
                    )
                ),
            )

            fig = go.Figure(data=data_input, layout=layout)
            if transpose:
                for i in val_sorter:
                    fig.add_vline(i, line_width=0.5)

                for i in between:
                    fig.add_vline(
                        i, line_dash="dash", line_color="lightgrey", layer="below"
                    )

                if mode == "samples":
                    for i in range(len(settings[value]["sample_index"]) + 1):
                        fig.add_hline(i - 0.5, line_width=0.5)
                else:
                    for i in range(len(settings[value]["condition_index"]) + 1):
                        fig.add_hline(i - 0.5, line_width=0.5)

                fig.add_vrect(
                    x0=0,
                    x1=val_sorter[len(experimental_factors[value])],
                    line=dict(color="red", width=5),
                    layer="above",
                )

            else:
                for i in val_sorter:
                    fig.add_hline(i, line_width=0.5)

                for i in between:
                    fig.add_hline(
                        i, line_dash="dash", line_color="lightgrey", layer="below"
                    )

                if mode == "samples":
                    for i in range(len(settings[value]["sample_index"]) + 1):
                        fig.add_vline(i - 0.5, line_width=0.5)
                else:
                    for i in range(len(settings[value]["condition_index"]) + 1):
                        fig.add_vline(i - 0.5, line_width=0.5)

                fig.add_hrect(
                    y0=0,
                    y1=val_sorter[len(experimental_factors[value])],
                    line=dict(color="red", width=5),
                    layer="above",
                )

            fig.update_layout(
                modebar={"remove": ["zoom", "pan", "zoomIn", "zoomOut", "autoScale"]}
            )
            fig.layout.yaxis.fixedrange = True

        heatmaps.append(
            (
                value,
                fig,
                conditions[value] if len(list(conditions[value].keys())) > 0 else None,
            )
        )
    return heatmaps


def color(color, text):
    return f"<span style='color:{str(color)}'> {str(text)} </span>"



import copy
import sys

sys.path.append("metadata-organizer")
import os
import random
import string
import subprocess
import time
import json

import fred.src.utils as utils
import fred.src.web_interface.editing as editing
import fred.src.web_interface.factors_and_conditions as fac_cond
import fred.src.web_interface.file_io as file_io
import fred.src.git_whitelists as gwi
import fred.src.web_interface.html_output as html_output
import fred.src.web_interface.searching as searching
import fred.src.web_interface.validation as validation
import fred.src.web_interface.whitelist_parsing as whitelist_parsing
import fred.src.web_interface.wi_object_to_yaml as oty
import fred.src.web_interface.yaml_to_wi_object as yto
import fred.src.heatmap.create_heatmap as create_heatmap
from jinja2 import Template

# This script contains all functions for generation of objects for the web
# interface


class Webinterface:

    def __init__(self, config):
        (
            self.whitelist_repo,
            self.whitelist_branch,
            self.whitelist_path,
            self.username,
            self.password,
            structure,
            self.update_whitelists,
            self.output_path,
            self.filename,
            self.email,
        ) = utils.parse_config(config)
        self.structure = utils.read_in_yaml(structure)
        self.whitelist_version = fetch_whitelists(self.__dict__)

    def to_dict(self):
        return self.__dict__


def fetch_whitelists(pgm_object):
    whitelist_version = gwi.get_whitelists(
        pgm_object["whitelist_path"],
        pgm_object["whitelist_repo"],
        pgm_object["whitelist_branch"],
        pgm_object["update_whitelists"],
    )
    return whitelist_version


def get_whitelist_object(pgm_object):
    whitelist_object = {
        "whitelists": whitelist_parsing.get_whitelist_object(pgm_object),
        "version": pgm_object["whitelist_version"],
    }
    return whitelist_object


def get_empty_wi_object(pgm_object, read_in_whitelists):
    return yto.get_empty_wi_object(pgm_object["structure"], read_in_whitelists)


def is_empty(pgm_object, wi_object, read_in_whitelsits):
    emtpy_object = yto.get_empty_wi_object(pgm_object["structure"], read_in_whitelsits)
    if wi_object == emtpy_object:
        empty = True
    else:
        empty = False
    return {"empty": empty, "object": emtpy_object}


def get_single_whitelist(ob, read_in_whitelists):
    return whitelist_parsing.get_single_whitelist(ob, read_in_whitelists)


def get_factors(pgm_object, organism, read_in_whitelists):
    return fac_cond.get_factors(organism, pgm_object["structure"], read_in_whitelists)


def get_conditions(pgm_object, factors, organism_name, read_in_whitelists):
    return fac_cond.get_conditions(
        factors, organism_name, pgm_object["structure"], read_in_whitelists
    )


def validate_object(pgm_object, wi_object, read_in_whitelists, finish=False):
    new_object = copy.deepcopy(wi_object)
    return validation.validate_object(
        new_object,
        pgm_object["structure"],
        read_in_whitelists,
        finish,
        pgm_object["email"],
    )


def get_summary(pgm_object, wi_object, read_in_whitelists):
    return html_output.get_summary(
        wi_object, pgm_object["structure"], read_in_whitelists
    )


def save_object(dictionary, path, filename, edit_state):
    object, id = file_io.save_object(dictionary, path, filename, edit_state)
    return object, id


def save_filenames(file_str, path):
    return file_io.save_filenames(file_str, path)


def get_plot(pgm_object, config=None, path=None, project_id=None, object=None):

    if object is not None:
        yaml_file = object
    elif config is not None and path is not None and project_id is not None:
        uuid = "".join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(5)
        )
        filename = f"{uuid}_{time.time()}"
        working_path = os.path.join(os.path.dirname(__file__), "..", "..")
        proc = subprocess.Popen(
            [
                "fred",
                "find",
                "-p",
                path,
                "-s",
                f'project:id:"{project_id}',
                "-c",
                config,
                "-o",
                "json",
                "-f",
                filename,
                "-sv",
            ],
            cwd=working_path,
        )
        proc.wait()
        res = utils.read_in_json(os.path.join(working_path, f"{filename}.json"))
        os.remove(os.path.join(working_path, f"{filename}.json"))
        try:
            yaml_file = utils.read_in_yaml(res["data"][0]["path"])
        except:
            yaml_file = None
    else: 
        yaml_file = None

    if yaml_file:
        try:
            template = Template(
            '''             
            {% if input.html %}
                {{ input.html }}
            {% else %}            
                <div style="overflow:auto; overflow-y:hidden; margin:0 auto; white-space:nowrap; padding-top:20">
                        {% if input.plot %}
                            {{ input.plot }}
                        {% endif %}
                        
                        {% if input.missing_samples %}
                            <i>Conditions without samples:</i>
                            {{ input.missing_samples }}
                        {% endif %}
                </div>
            {% endif %}
            '''
            )
            plots = create_heatmap.get_heatmap(
                yaml_file, pgm_object["structure"], show_setting_id=False
            )
            plot_list = []
            for elem in plots:
                add_plot = {}
                if elem[1] is not None:
                    add_plot["plot"] = elem[1]
                if elem[2] is not None:
                    add_plot["missing_samples"] = html_output.object_to_html(
                        elem[2], 0, False
                    )
                plot_list.append(
                    {"title": elem[0], "plot": template.render(input=add_plot)}
                )
        except:
            plot_list = []
    else:
        plot_list = []
    return plot_list

def download_plot(pgm_object, finished_yaml, save_path):
    plots = create_heatmap.get_heatmap(
                    finished_yaml, pgm_object["structure"], show_setting_id=True, labels="all", background=True
                )
    filenames = []
    try:
        project_id = finished_yaml['project']['id']
    except KeyError:
        project_id = 'missingID'
    for i in range(len(plots)):
        if plots[i][1] is not None:
            filename = f'{project_id}_{plots[i][0]}.png'
            plots[i][1].write_image(os.path.join(save_path, filename), format="png")
            filenames.append(filename)
    return filenames

def get_meta_info_from_object(object):
    try:
        project_id = object['project']['id']
    except KeyError:
        project_id = ''
    return searching.get_html_str('', object, project_id, None)


# TODO: fix path
def get_meta_info(config, path, project_ids):

    if not isinstance(project_ids, list):
        project_ids = [project_ids]

    metafile = {}
    html_str = ""
    for project_id in project_ids:
        uuid = "".join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(5)
        )
        filename = f"{uuid}_{time.time()}"
        working_path = os.path.join(os.path.dirname(__file__), "..", "..")
        proc = subprocess.Popen(
            [
                "fred",
                "find",
                "-p",
                path,
                "-s",
                f'project:id:"{project_id}',
                "-c",
                config,
                "-o",
                "json",
                "-f",
                filename,
                "-sv",
            ],
            cwd=working_path,
        )
        proc.wait()
        res = utils.read_in_json(os.path.join(working_path, f"{filename}.json"))
        os.remove(os.path.join(working_path, f"{filename}.json"))

        html_str, metafile = searching.get_meta_info(
            html_str,
            res["data"],
            project_id,
            res["validation_reports"] if "validation_reports" in res else None,
        )

    if html_str == "":
        html_str = "No metadata found.<br>"
    return html_str, metafile


def get_search_mask(pgm_object):
    return searching.get_search_mask(pgm_object["structure"])


# TODO: fix path
def find_metadata(config, path, search_string):
    start = time.time()
    uuid = "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(5)
    )
    filename = f"{uuid}_{time.time()}"
    working_path = os.path.join(os.path.dirname(__file__), "..", "..")
    proc = subprocess.Popen(
        [
            "fred",
            "find",
            "-p",
            path,
            "-s",
            search_string,
            "-c",
            config,
            "-o",
            "json",
            "-f",
            filename,
            "-sv",
        ],
        cwd=working_path,
    )
    proc.wait()
    subprocess_end = time.time()
    print(f'Subprocess "FIND" took {"%.2f" % (subprocess_end - start)} seconds.')
    res = utils.read_in_json(os.path.join(working_path, f"{filename}.json"))
    os.remove(os.path.join(working_path, f"{filename}.json"))
    read_end = time.time()
    print(
        f'Reading and removing the json file took {"%.2f" % (read_end - subprocess_end)} seconds.'
    )
    return res["data"]


def edit_wi_object(path, pgm_object, read_in_whitelists):
    return editing.edit_wi_object(path, pgm_object["structure"], read_in_whitelists)


# TODO: not needed -> in summary
def parse_object(pgm_object, wi_object, read_in_whitelists, return_id=False):
    # read in general structure
    return oty.parse_object(wi_object, pgm_object["structure"], read_in_whitelists, return_id=return_id)


def parse_search_string_to_query(search_string, structure):
    if not search_string.startswith("("):
        search_string = f'({search_string})'

    query = ""
    sub_list = ""
    sub = ""

    # set brace_count to 0 meaning that there is currently no open
    # brace
    brace_count = 0

    # iterate over search string
    for letter in search_string:

        # if an opening bracket is reached, add 1 to brace_count, add
        # the string within sub to the sub_list and set sub to an empty
        # string
        if letter == "(":
            brace_count += 1
            sub_list += sub
            sub = ""

        # if a closing bracket is reached, subtract 1 from brace_count
        elif letter == ")":
            brace_count -= 1

            if sub != "":
                res = parse_string_to_query_dict(sub, structure)
                sub_list += res

            # set sub to an empty string again
            sub = ""

        else:

            # add the current character of the search string to sub
            sub += letter

    sub_list = parse_string_to_query_dict(sub_list, structure)
    # add a dictionary containing information about the metadata file
    # to result if it matches the search
    # key: project id
    # value: whole metafile or path to metafile depending on whether
    # result_dict is set to True or False
    d = json.loads(sub_list)
    return d


def get_text_keys(structure, pre=''):
    keys =  []
    for key in structure:
        if 'input_type' in structure[key]:
            if structure[key]['input_type'] in ['short_text', 'long_text', 'select']:
                keys.append(f'{pre}.{key}'.strip('.'))
        else:
            if 'value' in structure[key] and isinstance(structure[key]['value'], dict):
                keys += get_text_keys(structure[key]['value'], key)
    return keys

def get_all_query(keys, value):
    key_query =  []
    for key in keys:
        key_query.append(f'{"{"} "{key}": {"{"} "$regex": "{value}", "$options": "i" {"}"} {"}"}')
    
    return key_query
    


def parse_string_to_query_dict(string, structure, all_keys=None):
    if ' or ' in string:
        or_vals = string.split(' or ')
        for i in range(len(or_vals)):
            if not or_vals[i].startswith('{ '):
                if 'and' in or_vals[i]:
                    and_vals = or_vals[i].split(' and ')
                    for j in range(len(and_vals)):
                        if not and_vals[j].startswith('{ '):
                            is_not = False
                            if and_vals[j].strip().startswith('not '):
                                and_vals[j] = and_vals[j].split('not ')[1].strip()
                                is_not = True
                            if '"' in and_vals[j]:
                                key, value = and_vals[j].rstrip('"').split('"')
                                if key != '':
                                    key = key.replace(':', '.').rstrip('.')
                                    value = f'{"{"} "$regex": "{value}", "$options": "i" {"}"}'
                                else:
                                    if all_keys is None:
                                        all_keys = get_text_keys(structure)
                                    key = "$or"
                                    value = f'[ {", ".join(get_all_query(all_keys, and_vals[j]))} ]'
                            else:
                                if all_keys is None:
                                    all_keys = get_text_keys(structure)
                                key = "$or"
                                value = f'[ {", ".join(get_all_query(all_keys, and_vals[j]))} ]'
                            if is_not:
                                and_vals[j] = f'{"{"} "$not": {"{"} "{key}": {value} {"}"} {"}"}'
                            else:
                                and_vals[j] = f'{"{"} "{key}": {value} {"}"}'
                    or_vals[i] = f'{"{"} "$and": [ {", ".join(and_vals)} ] {"}"}'
                else:
                    is_not = False
                    if or_vals[i].strip().startswith('not '):
                        or_vals[i] = or_vals[i].split('not ')[1].strip()
                        is_not = True
                    if '"' in or_vals[i]:
                        key, value = or_vals[i].rstrip('"').split('"')
                        if key != '':
                            key = key.replace(':', '.').rstrip('.')
                            value = f'{"{"} "$regex": "{value}", "$options": "i" {"}"}'
                        else:
                            if all_keys is None:
                                all_keys = get_text_keys(structure)
                            key = "$or"
                            value = f'[ {", ".join(get_all_query(all_keys, or_vals[i]))} ]'
                    else:
                        if all_keys is None:
                            all_keys = get_text_keys(structure)
                        key = "$or"
                        value = f'[ {", ".join(get_all_query(all_keys, or_vals[i]))} ]'
                    if is_not:
                        or_vals[i] = f'{"{"} "$not": {"{"} "{key}": {value} {"}"} {"}"}'
                    else:
                        or_vals[i] = f'{"{"} "{key}": {value} {"}"}'
        res = f'{"{"} "$or": [ {", ".join(or_vals)} ] {"}"}'
    elif ' and ' in string:
        and_vals = string.split(' and ')
        for i in range(len(and_vals)):
            if not and_vals[i].startswith('{ '):
                is_not = False
                if and_vals[i].strip().startswith('not '):
                    and_vals[i] = and_vals[i].split('not ')[1].strip()
                    is_not = True
                if '"' in and_vals[i]:
                    key, value = and_vals[i].rstrip('"').split('"')
                    if key != '':
                        key = key.replace(':', '.').rstrip('.') 
                        value = f'{"{"} "$regex": "{value}", "$options": "i" {"}"}'
                    else:
                        if all_keys is None:
                            all_keys = get_text_keys(structure)
                        key = "$or"
                        value = f'[ {", ".join(get_all_query(all_keys, and_vals[i]))} ]'
                else:
                    if all_keys is None:
                        all_keys = get_text_keys(structure)
                    key = "$or"
                    value = f'[ {", ".join(get_all_query(all_keys, and_vals[i]))} ]'
                if is_not:
                    and_vals[i] = f'{"{"} "$not": {"{"} "{key}": {value} {"}"} {"}"}'
                else:
                    and_vals[i] = f'{"{"} "{key}": {value} {"}"}'
        res = f'{"{"} "$and": [ {", ".join(and_vals)} ] {"}"}'
    else:
        is_not = False
        if not string.startswith('{ '):
            if string.strip().startswith('not '):
                string = string.split('not ')[1].strip()
                is_not = True
            if '"' in string:
                key, value = string.rstrip('"').split('"')
                if key != '':
                    key = key.replace(':', '.').rstrip('.') 
                    value = f'{"{"} "$regex": "{value}", "$options": "i" {"}"}'
                else:
                    if all_keys is None:
                        all_keys = get_text_keys(structure)
                    key = "$or"
                    value = f'[ {", ".join(get_all_query(all_keys, value))} ]'
            else:
                if all_keys is None:
                    all_keys = get_text_keys(structure)
                key = "$or"
                value = f'[ {", ".join(get_all_query(all_keys, string))} ]'
            if is_not:
                res = f'{"{"} "$not": {"{"} "{key}": {value} {"}"} {"}"}'
            else:
                res = f'{"{"} "{key}": {value} {"}"}'
        else:
            res=string
    return res
            

def get_metadata_search_view(metadata_path):

    metadata = utils.read_in_yaml(metadata_path)
    res = {}

    try:
        res["id"] = metadata["project"]["id"]
    except KeyError:
        res["id"] = None
    
    try:
        res["project_name"] = metadata["project"]["project_name"]
    except KeyError:
        res["project_name"] = None

    try:
        res["owner"] = metadata["project"]["owner"]["name"]
    except KeyError:
        res["owner"] = None

    try:
        res["email"] = metadata["project"]["owner"]["email"]
    except KeyError:
        res["email"] = None

    res["organisms"] = list(set(utils.find_keys(metadata, "organism_name")))

    technique = []

    techniques = list(utils.find_keys(metadata, "technique"))
    for elem in techniques:
        technique += elem
    technique = list(set(technique))
    res['technique'] = technique

    try:
        res["description"] = metadata["project"]["description"]
    except KeyError:
        res["description"] = None

    try:
        res["date"] = metadata["project"]["date"]
    except KeyError:
        res["date"] = None

    if "nerd" in metadata["project"]:
        nerds = []
        for nerd in metadata["project"]["nerd"]:
            nerds.append(nerd["name"])
        res["nerd"] = nerds
    else:
        res["nerd"] = None

    cell_type = list(set(utils.find_keys(metadata, "cell_type")))
    res["cell_type"] = cell_type

    tissue = []

    tissues = list(utils.find_keys(metadata, "tissue"))
    for elem in tissues:
        tissue += elem
    tissue = list(set(tissue))
    res["tissue"] = tissue

    # treatment
    treatment = []

    medical = list(
        utils.find_list_key(metadata, "medical_treatment:treatment_type")
    )
    treatment += list(set(medical))

    physical = list(utils.find_keys(metadata, "physical_treatment"))
    treatment += list(set(physical))

    injury = list(utils.find_list_key(metadata, "injury:injury_type"))
    treatment += list(set(injury))

    res["treatment"] = treatment

    # disease

    disease = list(utils.find_list_key(metadata, "disease:disease_type"))
    res["disease"] = list(set(disease))

    return res

def read_metadata(path):
    return utils.read_in_yaml(path)

def get_metadata(path):
    full_metadata = utils.read_in_yaml(path)
    search_view = get_metadata_search_view(path)
    return full_metadata, search_view


import copy

import fred.src.utils as utils
import fred.src.web_interface.wi_utils as wi_utils
import os


def parse_object(wi_object, key_yaml, read_in_whitelists, return_id=False):
    """
    This function parses a wi object back into a yaml
    :param key_yaml: the read in general structure
    :param wi_object: the filled wi object
    :return: result: a dictionary matching the metadata yaml structure
    """

    # define an empty dictionary to store the converted wi_object
    result = {}

    # TODO: needed? better solution?
    # save project id
    project_id = ""
    for elem in wi_object["project"]["input_fields"]:
        if elem["position"].split(":")[-1] == "id" and elem["value"] is not None:
            project_id = elem["value"]

    # set parameters organism, sample_name and nom -> get filled during
    # conversion
    organism = ""
    sample_name = ""
    nom = 1
    global_count = 1
    local_count = 1
    double = []
    publication_records = wi_object["publication_records"]
    # iterate over parts (from general structure to ensure order)
    for key in key_yaml:

        # make sure the key is present in the wi object
        if key in wi_object and key not in ["old_filenames", "old_sample_names"]:

            # parse every part into yaml format
            (
                result[key],
                organism,
                sample_name,
                nom,
                global_count,
                local_count,
                double,
            ) = parse_part(
                wi_object[key],
                key_yaml,
                copy.deepcopy(wi_object["all_factors"]),
                project_id,
                organism,
                sample_name,
                nom,
                global_count,
                local_count,
                double,
                read_in_whitelists,
                publication_records,
            )

    # remove keys with value None
    result = {k: v for k, v in result.items() if v is not None}

    sample_name_positions = get_sample_name_positions(result)
    for sample_pos in sample_name_positions:
        utils.fill_key(
            sample_pos,
            utils.create_sample_names(
                result,
                (
                    wi_object["old_sample_names"]
                    if "old_sample_names" in wi_object
                    else {}
                ),
                sample_pos,
                read_in_whitelists=read_in_whitelists,
            ),
            result,
        )

    filename_positions = get_filename_positions(result)
    for file_pos in filename_positions:
        utils.fill_key(
            file_pos,
            utils.create_filenames(
                result,
                double,
                file_pos,
                wi_object["old_filenames"] if "old_filenames" in wi_object else {},
                read_in_whitelists=read_in_whitelists,
            ),
            result,
        )

    if return_id:
        return (project_id, result)
    else:
        return result


def get_filename_positions(metafile):
    positions = [
        x + ["filenames"]
        for x in utils.get_key_positions(metafile, "technical_replicates", [], [])
    ]
    return positions


def get_sample_name_positions(metafile):
    positions = [
        x + ["sample_name"]
        for x in utils.get_key_positions(metafile, "technical_replicates", [], [])
    ]
    return positions


def split_name(elem, double, gene=True):
    new_name = []
    if "+" in elem:
        new_split = elem.split("+")
        for part in new_split:
            new_part, gene = split_name(part, double, gene=gene)
            if new_part != "":
                new_name.append(new_part)
        elem = "-".join(new_name)

    if "#" in elem:
        remove = elem.split("#")[0]
        elem, gene = split_name(elem[len(f"{remove}#") :], double, gene=gene)

    if "." in elem:
        if elem.lower() in [f"gn.{x.lower()}" for x in double]:
            gene = False
            elem = ""
        elif elem.lower().startswith("embl.") and gene is True:
            elem = ""
        else:
            elem = elem.split(".")[1]

    return elem, gene


def parse_part(
    wi_object,
    key_yaml,
    factors,
    project_id,
    organism,
    sample_name,
    nom,
    global_count,
    local_count,
    double,
    read_in_whitelists,
    publication_records,
):
    """
    This function parses a part of the wi object to create the yaml structure
    :param key_yaml: the read in general structure
    :param sample_name: the name of a sample build from condition name and
                        index of biological replicate
    :param wi_object: a part of the filled wi object
    :param factors: the selected experimental factors
    :param organism: the selected organism
    :param project_id: the project id
    :param nom: the number of measurements
    :return: val: the parsed part in yaml structure
             organism: the shortened version of the used organism
             sample_name: the name of the current sample
             nom: the number of measurements for the current sample
    """

    # initialize the converted value with None
    val = None

    # test if the object to parse is a dictionary
    if isinstance(wi_object, dict):

        # test if the values were stored in the 'list_value' key
        # (if the key takes a list or is of type single- or multi-autofill)
        if "list_value" in wi_object and not (
            "input_type" in wi_object
            and wi_object["input_type"] in ["single_autofill", "restricted_short_text"]
        ):

            # define an empty list to store the converted list values
            val = []
            is_cond = False
            # TODO: headers for list element (if gene is a list)
            # iterate over the list elements
            for i in range(len(wi_object["list_value"])):

                # test if the element is a dictionary
                if isinstance(wi_object["list_value"][i], dict):

                    # special case: condition
                    if (
                        wi_object["list_value"][i]["position"].split(":")[-1]
                        == "condition"
                    ):

                        # define emtpy list to save parsed samples to
                        samples = []
                        local_count = 1
                        is_cond = True
                        # iterate over all samples
                        for sub_elem in wi_object["list_value"][i]["list_value"]:

                            # convert samples
                            (
                                sample,
                                organism,
                                sample_name,
                                nom,
                                global_count,
                                local_count,
                                double,
                            ) = parse_list_part(
                                sub_elem,
                                key_yaml,
                                factors,
                                project_id,
                                organism,
                                sample_name,
                                nom,
                                global_count,
                                local_count,
                                double,
                                read_in_whitelists,
                                publication_records,
                            )

                            # remove empty keys
                            sample = {k: v for k, v in sample.items() if v is not None}

                            # add sample to the samples list
                            samples.append(sample)

                        # add dictionary with condition name and biological
                        # replicates (samples) to list parsed list
                        condition = {
                            "condition_name": wi_object["list_value"][i][
                                "correct_value"
                            ],
                            "biological_replicates": {"count": len(samples)},
                        }

                        # add samples to the condition dictionary if count > 0
                        if len(samples) > 0:
                            condition["biological_replicates"]["samples"] = samples

                        # add condition to list of converted values
                        val.append(condition)

                # test if list element is a list
                elif isinstance(wi_object["list_value"][i], list):

                    # special case: experimental setting
                    if wi_object["position"].split(":")[-1] == "experimental_setting":

                        # call parse function using the experimental factors
                        # with the same index as the setting
                        (
                            c_val,
                            organism,
                            sample_name,
                            nom,
                            global_count,
                            local_count,
                            double,
                        ) = parse_list_part(
                            wi_object["list_value"][i],
                            key_yaml,
                            factors[i],
                            project_id,
                            organism,
                            sample_name,
                            nom,
                            global_count,
                            local_count,
                            double,
                            read_in_whitelists,
                            publication_records,
                        )

                        # add the converted value to the list
                        val.append(c_val)

                    # no special case
                    else:

                        # call parse function with all experimental factors
                        (
                            c_val,
                            organism,
                            sample_name,
                            nom,
                            global_count,
                            local_count,
                            double,
                        ) = parse_list_part(
                            wi_object["list_value"][i],
                            key_yaml,
                            factors,
                            project_id,
                            organism,
                            sample_name,
                            nom,
                            global_count,
                            local_count,
                            double,
                            read_in_whitelists,
                            publication_records,
                        )

                        # add the converted value to the list
                        val.append(c_val)

                # if list element is str/int/bool
                else:

                    new_val = copy.deepcopy(wi_object["list_value"][i])
                    parse_w_key = None
                    if "whitelist_keys" in wi_object:
                        for w_key in wi_object["whitelist_keys"]:
                            if new_val.endswith(f"({w_key})"):
                                new_val = new_val.rstrip(f"({w_key})").strip()
                                parse_w_key = w_key
                                break

                    if "headers" in wi_object:
                        if isinstance(wi_object["headers"], dict):
                            if parse_w_key is not None:
                                if parse_w_key in wi_object["headers"]:
                                    new_val = wi_utils.parse_headers(
                                        wi_object["headers"][parse_w_key],
                                        new_val,
                                        mode="dict",
                                    )
                        else:
                            new_val = wi_utils.parse_headers(
                                wi_object["headers"], new_val, mode="dict"
                            )
                    # add list element to list
                    val.append(new_val)

            # set the converted value to None if it is an empty list
            if len(val) == 0:
                val = None
            elif is_cond:
                val = sorted(val, key=lambda x: x["condition_name"])

        # the values are not saved as a list
        else:

            # wi object contains input fields
            if "input_fields" in wi_object:

                # special case: technical replicates
                if wi_object["position"].split(":")[-1] == "technical_replicates":

                    # file_name = get_file_name(sample_name.removesuffix(f'_{sample_name.split("_")[-1]}'), double)
                    # TODO: comment
                    # t_sample_name = []
                    # t_file_name = []
                    count = [
                        x["value"]
                        for x in wi_object["input_fields"]
                        if x["position"].split(":")[-1] == "count"
                    ][0]

                    # for c in range(count):
                    #    for m in range(nom):
                    #        t_sample_name.append(f'{project_id}_'
                    #                             f'{organism}_'
                    #                             f'{sample_name}'
                    #                             f'_t{"{:02d}".format(c + 1)}_'
                    #                             f'm{"{:02d}".format(m + 1)}')
                    #        t_file_name.append(f'{project_id}__'
                    #                           f'{global_count}__'
                    #                           f'{file_name}__'
                    #                           f'{local_count}')
                    #        local_count += 1
                    #        global_count += 1

                    val = {"count": count}

                # no special case
                else:

                    # call this function on the input fields
                    (
                        val,
                        organism,
                        sample_name,
                        nom,
                        global_count,
                        local_count,
                        double,
                    ) = parse_part(
                        wi_object["input_fields"],
                        key_yaml,
                        factors,
                        project_id,
                        organism,
                        sample_name,
                        nom,
                        global_count,
                        local_count,
                        double,
                        read_in_whitelists,
                        publication_records,
                    )

            # no input fields
            else:

                # set the value that should be converted (saved in list_value
                # if input type is 'single_autofill', else saved in value)
                convert_value = (
                    wi_object["list_value"][0]
                    if wi_object["input_type"]
                    in ["single_autofill", "restricted_short_text"]
                    and len(wi_object["list_value"]) > 0
                    else wi_object["value"]
                )

                # test if value was filled
                if convert_value is not None:

                    # wi object contains whitelist keys
                    if "whitelist_keys" in wi_object:

                        # replace value with converted one
                        convert_value = wi_utils.parse_whitelist_keys(
                            wi_object["whitelist_keys"],
                            convert_value,
                            wi_object["headers"] if "headers" in wi_object else None,
                        )

                    # wi object contains headers but no whitelist keys
                    elif "headers" in wi_object:

                        # replace the original value with the one split
                        # according to the header
                        convert_value = wi_utils.parse_headers(
                            wi_object["headers"], convert_value
                        )

                    # value is of type value_unit
                    if wi_object["input_type"] == "value_unit":

                        # save the value and unit as a dictionary
                        val = {
                            "value": wi_object["value"],
                            "unit": wi_object["value_unit"],
                        }

                    # value is of type date
                    elif wi_object["input_type"] == "date":

                        # convert the date to a string of format 'DD.MM.YYYY'
                        val = wi_utils.date_to_str(wi_object["value"])

                    # value was changed for display -> original value saved at
                    # key 'correct_value'
                    elif "correct_value" in wi_object:

                        # special_case sample name
                        if wi_object["position"].split(":")[-1] == "sample_name":

                            # split and save index of sample from sample name
                            sample_count = int(convert_value.split("_")[-1])

                            # reformat sample name with ending 'b<index>' for
                            # number of biological replicate
                            val = (
                                f'{wi_object["correct_value"]}_b'
                                f'{"{:02d}".format(sample_count)}'
                            )

                            # set the sample name to the value
                            sample_name = val

                        else:

                            # save correct value
                            val = wi_object["correct_value"]

                    else:

                        if isinstance(convert_value, str):

                            info = list(
                                utils.find_keys(
                                    key_yaml, wi_object["position"].split(":")[-1]
                                )
                            )
                            if (
                                len(info) > 0
                                and "input_type" in info[0]
                                and info[0]["input_type"] == "bool"
                            ):
                                if convert_value == "True":
                                    val = True
                                else:
                                    val = False
                            else:
                                val = (
                                    convert_value.replace("\\n", " ")
                                    .replace("\n", " ")
                                    .replace("\\", " ")
                                )
                                val = " ".join(val.split())
                        else:
                            # save value
                            val = convert_value

                        # set the number of replicates
                        if (
                            wi_object["position"].split(":")[-1]
                            == "number_of_measurements"
                        ):
                            nom = val

    # wi object is a list
    elif isinstance(wi_object, list):

        # call parse list function
        val, organism, sample_name, nom, global_count, local_count, double = (
            parse_list_part(
                wi_object,
                key_yaml,
                factors,
                project_id,
                organism,
                sample_name,
                nom,
                global_count,
                local_count,
                double,
                read_in_whitelists,
                publication_records,
            )
        )

    return val, organism, sample_name, nom, global_count, local_count, double


def parse_list_part(
    wi_object,
    key_yaml,
    factors,
    project_id,
    organism,
    sample_name,
    nom,
    global_count,
    local_count,
    double,
    read_in_whitelists,
    publication_records,
):
    """
    This function parses a part of the wi object of type list into the yaml
    structure
    :param sample_name: the name of a sample build from condition name and
                        index of biological replicate
    :param key_yaml: the read in general structure
    :param wi_object: the part of the wi object
    :param factors: the selected experimental factors
    :param organism: the selected organism
    :param project_id: the project id
    :param nom: the number of measurements
    :return: res: the parsed part in yaml structure
             organism: the shortened version of the used organism
             sample_name: the name of the current sample
             nom: the number of measurements for the current sample
    """

    # create dictionary to save parsed elements to
    res = {}

    # iterate over list
    for i in range(len(wi_object)):

        val = None

        # special case: organism
        if wi_object[i]["position"].split(":")[-1] == "organism":

            # TODO: headers
            # save the organism name
            organism = wi_object[i]["value"].split(" ")[0]

            # read in the abbrev whitelist for organisms
            short = utils.get_whitelist(
                os.path.join("abbrev", "organism_name"),
                {"organism": organism},
                whitelist_object=read_in_whitelists,
            )["whitelist"]

            # save the shortened version of the organism
            organism = short[organism]

        # special case: experimental factors
        if wi_object[i]["position"].split(":")[-1] == "experimental_factors":

            # iterate over experimental factors
            for r in range(len(factors)):

                factors[r], double = parse_factor(factors[r], key_yaml, double)

            # set val to factors
            val = factors

        elif wi_object[i]["position"].split(":")[-1] == "publication":
            publications = []
            for elem in wi_object[i]["list_value"]:
                pubmed_id = None
                for sub_elem in elem:
                    if sub_elem["position"].split(":")[-1] == "pubmed_id":
                        pubmed_id = sub_elem["value"]
                        break
                if pubmed_id is not None:
                    if str(pubmed_id) not in publication_records:
                        publications.append({"pubmed_id": pubmed_id})
                    else:
                        pub_keys = list(utils.find_keys(key_yaml, "publication"))[0]
                        publication = {}
                        for key in pub_keys["value"]:
                            if key in publication_records[str(pubmed_id)]:
                                publication[key] = publication_records[str(pubmed_id)][
                                    key
                                ]
                        publications.append(publication)
            if len(publications) > 0:
                val = publications
        # no special case
        else:

            # call parse part function
            val, organism, sample_name, nom, global_count, local_count, double = (
                parse_part(
                    wi_object[i],
                    key_yaml,
                    factors,
                    project_id,
                    organism,
                    sample_name,
                    nom,
                    global_count,
                    local_count,
                    double,
                    read_in_whitelists,
                    publication_records,
                )
            )

        # test if the value is not empty
        if val is not None and (
            (type(val) in [str, list, dict] and len(val) > 0)
            or type(val) not in [str, list, dict]
        ):

            # overwrite the old value with the converted one
            res[wi_object[i]["position"].split(":")[-1]] = val

    return (
        res if len(res) > 0 else None,
        organism,
        sample_name,
        nom,
        global_count,
        local_count,
        double,
    )


def parse_factor(factors, key_yaml, double):

    # nested factor
    if len(factors["values"]) == 1 and isinstance(factors["values"][0], dict):

        if "nested_infos" in factors:
            for key_ in factors["nested_infos"]:
                if "double" in factors["nested_infos"][key_]:
                    double = list(set(double + factors["nested_infos"][key_]["double"]))

        # remove key 'multi'
        if "multi" in factors["values"][0]:
            factors["values"][0].pop("multi")

        # remove keys with None as value
        factors["values"][0] = {
            k: v for k, v in factors["values"][0].items() if v is not None
        }

        # test if the key in the 'values' dictionary matches the
        # factor and overwrite the dictionary with its list values
        # -> this is the case if a factor can be a list and can
        #    therefor occur multiple times in a condition
        # -> key multi has to be added -> change 'values' from list
        #    to dict
        # -> e.g. factor tissue -> values {tissue: [...], multi: }
        if list(factors["values"][0].keys()) == [factors["factor"]]:
            factors["values"] = factors["values"][0][factors["factor"]]

        for i in range(len(factors["values"])):
            if isinstance(factors["values"][i], dict):
                remove_keys = []
                for key in factors["values"][i]:

                    whitelist_keys = None
                    headers = None

                    if "nested_infos" in factors and key in factors["nested_infos"]:
                        if "whitelist_keys" in factors["nested_infos"][key]:
                            whitelist_keys = factors["nested_infos"][key][
                                "whitelist_keys"
                            ]
                        if "headers" in factors["nested_infos"][key]:
                            headers = factors["nested_infos"][key]["headers"]

                    if factors["values"][i][key] is None or (
                        isinstance(factors["values"][i][key], list)
                        and len(factors["values"][i][key]) == 0
                    ):
                        remove_keys.append(key)

                    if whitelist_keys is not None:

                        for j in range(len(factors["values"][i][key])):
                            factors["values"][i][key][j] = (
                                wi_utils.parse_whitelist_keys(
                                    whitelist_keys,
                                    factors["values"][i][key][j],
                                    headers,
                                    mode="dict",
                                )
                            )
                    elif headers is not None:
                        for j in range(len(factors["values"][i][key])):
                            factors["values"][i][key][j] = wi_utils.parse_headers(
                                headers, factors["values"][i][key][j], mode="dict"
                            )

                    # if isinstance(factors['values'][i][key], dict):
                    #    if not 'factor' in factors['values'][i][key]:
                    #        factors['values'][i][key]['factor'] = key
                    #    factors['values'][i][key] = parse_factor(factors['values'][i][key], key_yaml)
                    #    factors['values'][i][key].pop('factor')
                for r_key in remove_keys:
                    factors["values"][i].pop(r_key)

    else:

        # fetch the properties of the experimental factor from the
        # general structure
        infos = list(utils.find_keys(key_yaml, factors["factor"]))

        if "double" in factors:
            double = set(double + factors["double"])

        # iterate over values of experimental factor
        for j in range(len(factors["values"])):

            # factor contains whitelist keys
            if "whitelist_keys" in factors:

                # replace value with converted one
                factors["values"][j] = wi_utils.parse_whitelist_keys(
                    factors["whitelist_keys"],
                    factors["values"][j],
                    factors["headers"] if "headers" in factors else None,
                )

            # factor contains headers but no whitelist keys
            elif "headers" in factors:

                # replace the original value with the one split
                # according to the header
                factors["values"][j] = wi_utils.parse_headers(
                    factors["headers"], factors["values"][j]
                )

            # factor of type value_unit
            elif (
                len(infos) > 0
                and "special_case" in infos[0]
                and "value_unit" in infos[0]["special_case"]
            ):

                factors["values"][j] = utils.split_value_unit(factors["values"][j])

    # remove the keys 'header' and 'whitelist_keys'
    if "whitelist_keys" in factors:
        factors.pop("whitelist_keys")
    if "headers" in factors:
        factors.pop("headers")
    if "nested_infos" in factors:
        factors.pop("nested_infos")
    if "double" in factors:
        factors.pop("double")
    return factors, double



from fred.src.input_functions import Input
from fred.src.autogenerate import Autogenerate
from fred.src import validate_yaml
from fred.src import utils
from fred.src.heatmap import create_heatmap
import os
from jinja2 import Template

template = Template(
    """
        <h3>{{ header }}</h3>
                        
            {% for elem in plots %}

                {% if loop.index0 != 0 %}
                    <hr style="border-style: dotted;" />
                {% endif %}
                                
                <div style="woverflow:auto; overflow-y:hidden; margin:0 auto; white-space:nowrap; padding-top:5">
                    {{ elem.plot }}

                    {% if elem.missing_samples %}
                        <i>Conditions without samples:</i>
                        {{ elem.missing_samples }}
                    {% endif %}
                </div>
                                
            {% endfor %} 
                
        """
)


class Generate(Input):

    def generate(self):
        indent = 1
        for part in self.key_yaml:
            self.parse_lists(self.key_yaml[part], [part], indent, self.result_dict)
            if part == "experimental_setting":
                plot = create_heatmap.get_heatmap(
                    {part: self.result_dict[part]}, self.key_yaml
                )
                for elem in plot:
                    if elem[1] is not None:
                        elem[1].show()
            else:
                print(self.get_summary(self.result_dict[part]))
        for elem in self.generate_end:
            func = getattr(Autogenerate, f"get_{elem[-1]}")
            fill_val = func(Autogenerate(self, elem))
            if fill_val is not None:
                self.fill_key(elem, fill_val, self.result_dict)

        # print validation report
        # print(self.get_validation(self.result_dict))
        print(self.get_summary(self.result_dict))
        # save information to a yaml file and print the filename
        print(
            f"File is saved to "
            f'{os.path.join(self.path, f"{self.project_id}{self.filename}.yaml")}'
            f""
        )
        utils.save_as_yaml(
            self.result_dict,
            os.path.join(self.path, f"{self.project_id}{self.filename}.yaml"),
        )

    def print_sample_names(self):
        """
        This function creates a string out of all generated filenames that can
        be printed.
        """
        samples = list(
            utils.find_list_key(self.result_dict, "technical_replicates:sample_name")
        )
        print(
            f'{"".center(self.size, "-")}\n'
            f'{"SAMPLE NAMES".center(self.size, " ")}\n'
            f'{"".center(self.size, "-")}\n'
        )
        sample_names = ""
        for elem in samples:
            for name in elem:
                sample_names += f"- {name}\n"
        print(sample_names)
        save = self.parse_list_choose_one(
            ["True ", "False "], "Do you want to save the sample names into a file?"
        )
        if save:
            text_file = open(
                os.path.join(self.path, f"{self.project_id}_samples.txt"), "w"
            )
            text_file.write(sample_names)
            text_file.close()
            print(
                f"The sample names have been saved to file "
                f"'{self.path}/{self.project_id}_samples.txt'."
            )

    def get_summary(self, result):
        summary = ""
        summary += (
            f'{"".center(self.size, "=")}\n'
            f'{"SUMMARY".center(self.size, " ")}\n'
            f'{"".center(self.size, "=")}\n'
        )
        summary += self.print_summary(result, 1, False)
        summary += f"\n\n"
        summary += f'{"".center(self.size, "=")}\n'
        return summary

    def get_validation(self, result):
        validation_reports = {
            "all_files": 1,
            "corrupt_files": {"count": 0, "report": []},
            "error_count": 0,
            "warning_count": 0,
        }
        file_reports = {"file": result, "error": None, "warning": None}
        report = ""
        report += (
            f'{"FILE VALIDATION".center(self.size, " ")}\n'
            f'{"".center(self.size, "-")}\n'
        )
        (
            valid,
            missing_mandatory_keys,
            invalid_keys,
            invalid_entries,
            invalid_values,
            logical_warn,
        ) = validate_yaml.validate_file(result, self.key_yaml, self.filename)
        if not valid:
            validation_reports["corrupt_files"]["count"] = 1
            validation_reports["error_count"] += (
                len(missing_mandatory_keys)
                + len(invalid_keys)
                + len(invalid_entries)
                + len(invalid_values)
            )
            file_reports["error"] = (
                missing_mandatory_keys,
                invalid_keys,
                invalid_entries,
                invalid_values,
            )
        if len(logical_warn) > 0:
            validation_reports["corrupt_files"]["count"] = 1
            validation_reports["warning_count"] += len(logical_warn)
            file_reports["warning"] = logical_warn
        validation_reports["corrupt_files"]["report"].append(file_reports)

        report += (
            f'Found {validation_reports["error_count"]} errors and '
            f'{validation_reports["warning_count"]} warnings.\n'
        )

        if validation_reports["corrupt_files"]["count"] > 0:
            rep = ""
            for _report in validation_reports["corrupt_files"]["report"]:
                rep += f'{"".center(self.size, "_")}\n\n'
                rep += validate_yaml.print_full_report(
                    _report["file"], _report["error"], _report["warning"], self.size
                )
            rep += f'{"".center(self.size, "_")}\n\n'
            report += rep
        return report

    def print_summary(self, result, depth, is_list):
        """
        This function parses the dictionary into a string with the same
        structure as it will be saved to the yaml file
        :param result: the filled dictionary
        :param depth: an integer that specifies the depth of indentation
        :param is_list: a bool that states if a key contains a list
        :return: summary: a string that contains all entered information
        """
        summary = ""
        if isinstance(result, dict):
            for key in result:
                printed_summary = self.print_summary(result[key], depth + 1, is_list)
                if key == list(result.keys())[0] and is_list:
                    summary = (
                        f'{summary}\n{"    " * (depth - 1)}{"  - "}'
                        f"{key}: {printed_summary}"
                    )
                else:
                    summary = f'{summary}\n{"    " * depth}{key}: ' f"{printed_summary}"
        elif isinstance(result, list):
            for elem in result:
                if not isinstance(elem, list) and not isinstance(elem, dict):
                    summary = f'{summary}\n{"    " * (depth - 1)}{"  - "}' f"{elem}"
                else:
                    summary = f"{summary}" f"{self.print_summary(elem, depth, True)}"
        else:
            summary = f"{summary}{result}"
        return summary

    def parse_lists(self, structure, position, indent, return_dict, is_factor=False):

        if isinstance(structure["value"], dict):
            elem_index = 0
            redo = True

            while redo:
                self.print_header(structure["display_name"], indent)
                print(utils.print_desc(structure["desc"], size=self.size), "\n")
                is_list = False
                if "special_case" in structure and "merge" in structure["special_case"]:
                    self.fill_key(
                        (
                            position + [elem_index]
                            if structure["list"] or is_factor
                            else position
                        ),
                        self.parse_input_value(
                            position[-1],
                            structure["value"][structure["special_case"]["merge"]],
                        ),
                        return_dict,
                    )
                    if is_factor:
                        is_list = True
                elif (
                    "special_case" in structure
                    and "value_unit" in structure["special_case"]
                ):
                    if structure["list"] or is_factor:
                        self.fill_key(
                            position,
                            self.get_list_value_unit(position[-1], structure),
                            return_dict,
                        )
                    else:
                        self.fill_key(
                            position, self.get_value_unit(structure), return_dict
                        )
                else:
                    self.input_keys(
                        structure["value"],
                        (
                            position + [elem_index]
                            if structure["list"] or is_factor
                            else position
                        ),
                        indent,
                        return_dict,
                        is_factor=is_factor,
                    )
                self.print_headline(indent)
                if (structure["list"] and not is_factor) or (is_list and is_factor):
                    redo = self.parse_list_choose_one(
                        ["True ", "False "],
                        f"\nDo you want to add another "
                        f'{structure["display_name"]}?',
                    )
                    elem_index += 1
                else:
                    redo = False

        else:
            self.fill_key(
                position,
                (
                    self.get_input_list(structure, position[-1])
                    if structure["list"] or is_factor
                    else self.parse_input_value(position[-1], structure)
                ),
                return_dict,
            )

    def print_headline(self, indent):
        if indent > 1:
            delim = "-"
        else:
            delim = "_"
        print(f'\n{"".center(self.size, delim)}\n')

    def print_header(self, key, indent):
        if indent > 1:
            delim = "-"
            new_line = ""
        else:
            delim = "_"
            new_line = "\n"
        print(
            f"\n"
            f'{"".center(self.size, delim)}{new_line}\n'
            f'{f"{key}".center(self.size, " ")}\n'
            f'{"".center(self.size, delim)}\n'
        )

    def input_keys(self, structure, position, indent, return_dict, is_factor=False):
        optionals = []
        desc = []
        for key in structure:
            if structure[key]["mandatory"] or (
                "special_case" in structure[key]
                and (
                    "generated" in structure[key]["special_case"]
                    or "factor" in structure[key]["special_case"]
                )
            ):
                if "special_case" in structure[key]:
                    if (
                        "factor" in structure[key]["special_case"]
                        and structure[key]["special_case"]["factor"]
                    ):
                        if "list" in structure[key] and structure[key]["list"]:
                            print(structure[key]["value"])
                            self.fill_key(
                                position + [key], [structure[key]["value"]], return_dict
                            )
                        else:
                            self.fill_key(
                                position + [key], structure[key]["value"], return_dict
                            )
                    elif "generated" in structure[key]["special_case"]:
                        if structure[key]["special_case"]["generated"] == "now":
                            func = getattr(Autogenerate, f"get_{key}")
                            fill_val = func(Autogenerate(self, position + [key]))
                            if fill_val is not None:
                                self.fill_key(position + [key], fill_val, return_dict)
                                if not isinstance(structure[key]["value"], dict):
                                    print(f'\n---\n{structure[key]["desc"]}\n')
                                    print(
                                        f"{key}: {utils.find_position(self.result_dict, position + [key])}"
                                    )
                        elif structure[key]["special_case"]["generated"] == "end":
                            self.generate_end.append(position + [key])
                        elif structure[key]["special_case"]["generated"] == "fill":
                            self.fill_key(
                                position + [key], structure[key]["value"], return_dict
                            )
                            optionals.append(key)
                            desc.append(structure[key]["desc"])
                    elif "value_unit" in structure[key]["special_case"]:
                        if "list" in structure[key]:
                            self.fill_key(
                                position + [key],
                                self.get_list_value_unit(key, structure),
                                return_dict,
                            )
                        else:
                            self.fill_key(
                                position + [key],
                                self.get_value_unit(structure),
                                return_dict,
                            )
                    elif "merge" in structure[key]["special_case"]:
                        self.fill_key(
                            position + [key],
                            self.parse_input_value(
                                key,
                                structure[key]["value"][
                                    structure[key]["special_case"]["merge"]
                                ],
                            ),
                            return_dict,
                        )
                    else:
                        self.parse_lists(
                            structure[key],
                            position + [key],
                            indent + 1,
                            return_dict,
                            is_factor=is_factor,
                        )
                else:
                    self.parse_lists(
                        structure[key],
                        position + [key],
                        indent + 1,
                        return_dict,
                        is_factor=is_factor,
                    )
            else:
                if (
                    not (
                        "special_case" in structure[key]
                        and "factor" in structure[key]["special_case"]
                        and structure[key]["special_case"]["factor"]
                    )
                    or structure[key]["list"]
                ):
                    optionals.append(key)
                    desc.append(structure[key]["desc"])
        if len(optionals) > 0 and not self.mandatory_only:
            print(
                f"\nDo you want to add any of the following optional keys?"
                f" (1,...,{len(optionals)} or n)\n"
            )
            self.print_option_list(optionals, desc)
            options = self.parse_input_list(optionals, True)
            if options:
                for option in options:
                    if (
                        "special_case" in structure[option]
                        and "generated" in structure[option]["special_case"]
                        and structure[option]["special_case"]["generated"] != "fill"
                    ):
                        if structure[option]["special_case"]["generated"] == "now":
                            func = getattr(Autogenerate, f"get_{option}")
                            fill_val = func(Autogenerate(self, position + [option]))
                            if fill_val is not None:
                                self.fill_key(
                                    position + [option], fill_val, return_dict
                                )
                        elif structure[option]["special_case"]["generated"] == "end":
                            self.generate_end.append(position + [option])
                    else:
                        self.parse_lists(
                            structure[option],
                            position + [option],
                            indent + 1,
                            return_dict,
                            is_factor=is_factor,
                        )

    def fill_key(self, position, value, fill_dict):
        if len(position) > 0:
            if len(position) == 1:
                fill_dict[position[0]] = value
            else:
                if type(position[1]) == str:
                    if type(position[0]) == str and position[0] not in fill_dict:
                        fill_dict[position[0]] = {}
                else:
                    if position[0] not in fill_dict:
                        fill_dict[position[0]] = []
                    if len(fill_dict[position[0]]) < position[1] + 1:
                        for i in range(len(fill_dict[position[0]]), position[1] + 1):
                            fill_dict[position[0]].append({})
                self.fill_key(position[1:], value, fill_dict[position[0]])
        else:
            print("NO POSITION")



import argparse
import copy
import pathlib
import sys
import os
import time

sys.path.append(os.path.dirname(__file__))
from fred.src.generate import Generate
from fred.src import find_metafiles
from fred.src import validate_yaml
from fred.src import file_reading
from fred.src import utils
from fred.src import git_whitelists
from fred.src.heatmap import create_heatmap


class FRED:

    def __init__(self, config):
        (
            self.whitelist_repo,
            self.whitelist_branch,
            self.whitelist_path,
            self.username,
            self.password,
            structure,
            self.update_whitelists,
            self.output_path,
            self.filename,
            self.email,
        ) = utils.parse_config(config)
        self.fetch_whitelists()
        self.structure = utils.read_in_yaml(structure)

    def fetch_whitelists(self):
        git_whitelists.get_whitelists(
            self.whitelist_path,
            self.whitelist_repo,
            self.whitelist_branch,
            self.update_whitelists,
        )

    def find(self, search_path, search, output, output_filename, skip_validation):
        result = find_metafiles.find_projects(
            self.structure, search_path, search, True, skip_validation
        )
        if output == "print":
            if len(result) > 0:

                # print summary of matching files
                print(find_metafiles.print_summary(result, output))

            else:

                # print information that there are no matching files
                print("No matches found")
        elif output == "json":
            if not output_filename:
                output_filename = "search_result"
            json_filename = f"{output_filename}.json"
            utils.save_as_json(
                {"data": find_metafiles.print_summary(result, output)}, json_filename
            )
            print(f"The report was saved to the file '{json_filename}'.")

    def generate(self, path, project_id, mandatory_only):
        gen = Generate(
            path, project_id, mandatory_only, self.filename, self.structure, self.email
        )
        gen.generate()

    def validate(
        self, logical_validation, path, output, output_filename, save_empty=False
    ):
        validation_reports = {
            "all_files": 1,
            "corrupt_files": {"count": 0, "report": []},
            "error_count": 0,
            "warning_count": 0,
        }
        if os.path.isdir(path):
            metafiles, validation_reports = file_reading.iterate_dir_metafiles(
                self.structure,
                [path],
                filename=self.filename,
                logical_validation=logical_validation,
                yaml=copy.deepcopy(self.structure),
                whitelist_path=self.whitelist_path,
            )
        else:
            metafile = utils.read_in_yaml(path)
            file_reports = {"file": metafile, "error": None, "warning": None}
            (
                valid,
                missing_mandatory_keys,
                invalid_keys,
                invalid_entries,
                invalid_values,
                logical_warn,
            ) = validate_yaml.validate_file(
                metafile,
                self.structure,
                self.filename,
                logical_validation=logical_validation,
                yaml=copy.deepcopy(self.structure),
                whitelist_path=self.whitelist_path,
            )
            metafile["path"] = path
            if not valid:
                validation_reports["corrupt_files"]["count"] = 1
                validation_reports["error_count"] += (
                    len(missing_mandatory_keys)
                    + len(invalid_keys)
                    + len(invalid_entries)
                    + len(invalid_values)
                )
                file_reports["error"] = (
                    missing_mandatory_keys,
                    invalid_keys,
                    invalid_entries,
                    invalid_values,
                )
            if len(logical_warn) > 0:
                validation_reports["corrupt_files"]["count"] = 1
                validation_reports["warning_count"] += len(logical_warn)
                file_reports["warning"] = logical_warn
            validation_reports["corrupt_files"]["report"].append(file_reports)

        print(f'{validation_reports["all_files"]} files were validated.')
        print(
            f'Found {validation_reports["error_count"]} errors and {validation_reports["warning_count"]} warnings in {validation_reports["corrupt_files"]["count"]} of those files.'
        )

        if validation_reports["corrupt_files"]["count"] > 0 or save_empty is True:

            res = None
            if output is not None:
                if output == "print":
                    res = ["print report"]
                elif output == "txt":
                    res = ["save report to txt file"]
                elif output == "json":
                    res = ["save report to json file"]
                elif output == "yaml":
                    res = ["save report to yaml file"]
            else:
                options = [
                    "print report",
                    "save report to txt file",
                    "save report to json file",
                    "save report to yaml file",
                ]
                print(
                    f"Do you want to see a report? Choose from the following options (1,...,{len(options)} or n)"
                )
                ask = Generate("", "", False, self.filename, self.structure, self.email)
                ask.print_option_list(options, "")
                res = ask.parse_input_list(options, True)

            try:
                output_report = {
                    "report": copy.deepcopy(validation_reports)["corrupt_files"][
                        "report"
                    ]
                }
            except KeyError:
                output_report = {"report": []}
            for elem in output_report["report"]:
                id = list(utils.find_keys(elem["file"], "id"))
                if len(id) > 0:
                    elem["id"] = id[0]
                else:
                    elem["id"] = "missing"
                elem["path"] = elem["file"]["path"]
                errors = (
                    list(elem["error"])
                    if "error" in elem and elem["error"] is not None
                    else []
                )
                elem["error"] = {}
                elem.pop("file")
                for i in range(len(errors)):

                    if len(errors[i]) > 0:
                        if i == 0:
                            elem["error"]["missing_mandatory_keys"] = errors[i]
                        elif i == 1:
                            elem["error"]["invalid_keys"] = errors[i]
                        elif i == 2:
                            whitelist_values = []
                            for v in errors[i]:
                                key = ":".join(v.split(":")[:-1])
                                entry = v.split(":")[-1]
                                whitelist_values.append(entry + " in " + key + "\n")
                            elem["error"]["invalid_entries"] = whitelist_values
                        elif i == 3:
                            value = []
                            for v in errors[i]:
                                value.append(f"{v[0]}: {v[1]} -> {v[2]}")
                            elem["error"]["invalid_values"] = value

                if "warning" in elem:
                    if elem["warning"] is not None:
                        for i in range(len(elem["warning"])):
                            elem["warning"][
                                i
                            ] = f'{elem["warning"][i][0]}: {elem["warning"][i][1]}'
                    else:
                        elem.pop("warning")

            if res is not None:
                if output_filename is None:
                    timestamp = time.time()
                    output_filename = (
                        f'validation_report_{str(timestamp).split(".")[0]}'
                    )

                rep = ""
                for report in validation_reports["corrupt_files"]["report"]:
                    rep += f'{"".center(80, "_")}\n\n'
                    rep += validate_yaml.print_full_report(
                        report["file"], report["error"], report["warning"]
                    )
                rep += f'{"".center(80, "_")}\n\n'

                if "save report to txt file" in res:
                    txt_filename = f"{output_filename}.txt"
                    txt_f = open(txt_filename, "w")
                    txt_f.write(rep)
                    print(f"The report was saved to the file '{txt_filename}'.")
                    txt_f.close()

                if "save report to json file" in res:
                    json_filename = f"{output_filename}.json"
                    utils.save_as_json(output_report, json_filename)
                    print(f"The report was saved to the file '{json_filename}'.")

                if "save report to yaml file" in res:
                    yaml_filename = f"{output_filename}.yaml"
                    utils.save_as_yaml(output_report, yaml_filename)
                    print(f"The report was saved to the file '{yaml_filename}'.")

                if "print report" in res:
                    print(rep)

        return validation_reports["error_count"], validation_reports["warning_count"]

    # def edit(self, path, mandatory_only):
    #    try:
    #        size = os.get_terminal_size()
    #        size = size.columns
    #    except OSError:
    #        size = 80

    #    edit_file.edit_file(path, self.filename, mandatory_only, size)

    def add_value(self, path, position, value, edit_existing):
        files, errors = file_reading.iterate_dir_metafiles(
            self.structure,
            [path],
            self.filename,
            False,
            return_false=True,
            whitelist_path=self.whitelist_path,
        )
        position = position.split(":")
        # TODO: type
        for file in files:
            file = utils.add_value_at_pos(
                self.structure, file, position, value, edit_existing
            )
            save_path = file["path"]
            file.pop("path")
            print(f"edited file {save_path}")
            utils.save_as_yaml(file, save_path)


def find(args):
    """
    calls script find_metafiles to find matching files and print results
    :param args:
        path: a path of a folder that should be searched for metadata files
        search: a string specifying search parameters linked via 'and', 'or'
                and 'not'
    """
    finding = FRED(args.config)
    finding.find(
        args.path, args.search, args.output, args.filename, args.skip_validation
    )


def generate(args):
    """
    calls script generate_metafile to start dialog
    :param args:
    """
    generating = FRED(args.config)
    generating.generate(args.path, args.id, args.mandatory_only)


def validate(args):
    validating = FRED(args.config)
    errors, warnings = validating.validate(
        not args.skip_logic, args.path, args.output, args.filename
    )


def plot(args):
    fred_object = FRED(args.config)
    input_file = utils.read_in_yaml(args.path)
    plots = create_heatmap.get_heatmap(
        input_file,
        fred_object.structure,
        mode=args.mode,
        labels=args.labels,
        background=args.background,
        sample_labels=args.sample_labels,
        condition_labels=args.condition_labels,
        transpose=args.transpose,
        drop_defaults=args.drop_defaults,
    )
    output_filename = args.filename if args.filename is not None else "fig1"
    if len(plots) > 0:
        try:
            plot = plots[args.setting - 1][1]
        except IndexError:
            print(f"Setting exp{args.setting} does not exist. Defaulting to exp1.")
            plot = plots[0][1]

        if plot is not None:
            if args.output == "png":
                plot.write_image(f"{output_filename}.{args.output}", format="png")
                print(f"Plot was saved to {output_filename}.{args.output}")
            elif args.output == "html":
                with open(f"{output_filename}.{args.output}", "w") as file:
                    file.write(plot.to_html(full_html=False, include_plotlyjs="cdn"))
                print(f"Plot was saved to {output_filename}.{args.output}")
            else:
                plot.show()
        else:
            print("Plot could not be created due to missing samples or conditions.")
    else:
        print("No settings found.")


# def edit(args):
#    editing = FRED(args.config)
#    editing.edit(args.path, args.mandatory_only)


def add_value(args):
    adding = FRED(args.config)
    adding.add_value(args.path, args.position, args.value, args.edit_existing)


def main():

    parser = argparse.ArgumentParser(prog="metaTools.py")
    subparsers = parser.add_subparsers(title="commands")

    find_function = subparsers.add_parser(
        "find",
        help="This command is used to find "
        "projects by searching the "
        "metadata files.",
    )

    find_group = find_function.add_argument_group("mandatory arguments")
    find_group.add_argument(
        "-p", "--path", type=pathlib.Path, required=True, help="The path to be searched"
    )
    find_group.add_argument(
        "-s", "--search", type=str, required=True, help="The search parameters"
    )
    find_group.add_argument(
        "-c",
        "--config",
        type=pathlib.Path,
        help="Config file",
        default=os.path.join(os.path.dirname(__file__), "config", "config.yaml"),
    )
    find_group.add_argument(
        "-o", "--output", default="print", choices=["json", "print"]
    )
    find_group.add_argument("-f", "--filename", default=None)
    find_group.add_argument(
        "-sv", "--skip_validation", default=False, action="store_true"
    )
    find_function.set_defaults(func=find)

    create_function = subparsers.add_parser(
        "generate", help="This command is used to " "create a metadata file."
    )
    create_group = create_function.add_argument_group("mandatory arguments")
    create_group.add_argument(
        "-p",
        "--path",
        type=pathlib.Path,
        required=True,
        help="The path to save the yaml",
    )
    create_group.add_argument(
        "-c",
        "--config",
        type=pathlib.Path,
        help="Config file",
        default=os.path.join(os.path.dirname(__file__), "config", "config.yaml"),
    )
    create_group.add_argument(
        "-id", "--id", type=str, required=True, help="The ID of the experiment"
    )
    create_function.add_argument(
        "-mo",
        "--mandatory_only",
        default=False,
        action="store_true",
        help="If True, only mandatory keys will " "be filled out",
    )
    create_function.add_argument(
        "-m", "--mode", default="metadata", choices=["metadata", "mamplan"]
    )
    create_function.set_defaults(func=generate)

    validate_function = subparsers.add_parser("validate", help="")
    validate_group = validate_function.add_argument_group("mandatory arguments")
    validate_group.add_argument("-p", "--path", type=pathlib.Path, required=True)
    validate_function.add_argument(
        "-l", "--skip_logic", default=False, action="store_true"
    )
    validate_function.add_argument(
        "-m", "--mode", default="metadata", choices=["metadata", "mamplan"]
    )
    validate_function.add_argument(
        "-o", "--output", default=None, choices=["json", "txt", "print", "yaml"]
    )
    validate_function.add_argument("-f", "--filename", default=None)
    validate_function.add_argument(
        "-c",
        "--config",
        type=pathlib.Path,
        help="Config file",
        default=os.path.join(os.path.dirname(__file__), "config", "config.yaml"),
    )
    validate_function.set_defaults(func=validate)

    # edit_function = subparsers.add_parser('edit', help='')
    # edit_group = edit_function.add_argument_group('mandatory_arguments')
    # edit_group.add_argument('-p', '--path', type=pathlib.Path, required=True)
    # edit_function.add_argument('-mo', '--mandatory_only', default=False,
    #                             action='store_true',
    #                             help='If True, only mandatory keys will '
    #                                  'be filled out')
    # edit_function.add_argument('-c', '--config', type=pathlib.Path,
    #                          help='Config file', default='config.yaml')
    # edit_function.add_argument('-m', '--mode', default='metadata', choices=['metadata', 'mamplan'])
    # edit_function.set_defaults(func=edit)

    add_value_function = subparsers.add_parser("add_value", help="")
    add_value_function.add_argument(
        "-m", "--mode", default="metadata", choices=["metadata", "mamplan"]
    )
    add_value_function.add_argument("-pos", "--position", required=True)
    add_value_function.add_argument("-v", "--value", required=True)
    add_value_function.add_argument(
        "-t", "--type", default="str", choices=["str", "int", "float", "bool"]
    )
    add_value_function.add_argument("-p", "--path", type=pathlib.Path, required=True)
    add_value_function.add_argument(
        "-e", "--edit_existing", default=False, action="store_true"
    )
    add_value_function.add_argument(
        "-c",
        "--config",
        type=pathlib.Path,
        help="Config file",
        default=os.path.join(os.path.dirname(__file__), "config", "config.yaml"),
    )
    add_value_function.set_defaults(func=add_value)

    plot_function = subparsers.add_parser("plot", help="")
    plot_function.add_argument("-p", "--path", type=pathlib.Path, required=True)
    plot_function.add_argument(
        "-c",
        "--config",
        type=pathlib.Path,
        help="Config file",
        default=os.path.join(os.path.dirname(__file__), "config", "config.yaml"),
    )
    plot_function.add_argument(
        "-m", "--mode", default="samples", choices=["samples", "conditions"]
    )
    plot_function.add_argument("-s", "--setting", type=int, default=1)
    plot_function.add_argument(
        "-l", "--labels", default="factors", choices=["factors", "all", "none"]
    )
    plot_function.add_argument(
        "-o", "--output", default="show", choices=["show", "png", "html", "dash"]
    )
    plot_function.add_argument("-f", "--filename", type=pathlib.Path)
    plot_function.add_argument(
        "-b",
        "--background",
        default=False,
        action="store_true",
        help="If stated, the background will be displayed in white. Per default it is transparent.",
    )
    plot_function.add_argument(
        "-cl",
        "--condition_labels",
        default=False,
        action="store_true",
        help="If stated, the label of the condition will be displayed as a name. Per default an index is stated. ",
    )
    plot_function.add_argument(
        "-sl",
        "--sample_labels",
        default=False,
        action="store_true",
        help="If stated, the label of the sample will be displayed as a name. Per default an index is stated. ",
    )
    plot_function.add_argument(
        "-t",
        "--transpose",
        default=False,
        action="store_true",
    )
    plot_function.add_argument(
        "-d",
        "--drop_defaults",
        default=False,
        action="store_true",
    )
    plot_function.set_defaults(func=plot)

    args = parser.parse_args()

    try:
        args.func(args)
    except AttributeError:
        parser.print_help()


if __name__ == "__main__":

    main()



import fred.src.validate_yaml as validate_yaml
import fred.src.web_interface.wi_utils as wi_utils
import fred.src.web_interface.wi_object_to_yaml as oty
import fred.src.utils as utils

# TODO: rewrite and documentation


def validate_object(wi_object, key_yaml, read_in_whitelists, finish, email):
    """
    This function performs a validation over the wi object
    :param wi_object: the filled wi object
    :return: validation_object: the validated wi object with errors and
                                warnings
    """
    pooled = None
    organisms = []
    settings = []
    for setting in wi_object["experimental_setting"]["list_value"]:
        for elem in setting:
            if elem["position"].split(":")[-1] == "organism":
                if isinstance(elem["value"], dict):
                    organisms.append(elem["value"]["organism_name"].split(" ")[0])
                else:
                    organisms.append(elem["value"].split(" ")[0])
            elif elem["position"].split(":")[-1] == "setting_id":
                settings.append(elem["value"])
    warnings = {}
    errors = {}
    pubmed_ids = []

    for part in ["project", "experimental_setting", "technical_details"]:

        (
            part,
            wi_object[part],
            pooled,
            organisms,
            warnings[part],
            errors[part],
            pubmed_ids,
        ) = validate_part(
            part, wi_object[part], [], pooled, organisms, settings, [], pubmed_ids
        )

    pubmed_ids = list(set(pubmed_ids))
    if any(x not in wi_object["publication_records"] for x in pubmed_ids):
        wi_object["publication_records"] = get_publications(
            wi_object["publication_records"], pubmed_ids, email
        )

    if finish:
        parsed = oty.parse_object(wi_object, key_yaml, read_in_whitelists)
        for part in ["project", "experimental_setting", "technical_details"]:
            if part not in parsed:
                parsed[part] = {}
            (
                valid,
                missing_mandatory_keys,
                invalid_keys,
                invalid_entries,
                invalid_value,
                logical_warn,
            ) = validate_yaml.validate_file(
                parsed[part],
                key_yaml[part]["value"],
                "metadata",
                logical_validation=False,
                generated=False,
                only_mandatory=True,
            )
            for elem in missing_mandatory_keys:
                diaplay_name = list(utils.find_keys(key_yaml, elem.split(":")[-1]))[0][
                    "display_name"
                ]
                errors[part].append(
                    {
                        "position": f"{part}:{elem}",
                        "title": f"Missing {diaplay_name}",
                        "message": "Missing mandatory input",
                    }
                )
    validation_object = {"object": wi_object, "errors": errors, "warnings": warnings}
    return validation_object


def get_publications(publications, pubmed_ids, email):
    pubmed_entries = utils.get_publication_object(pubmed_ids, email)
    for pubmed_entry in pubmed_entries:
        if pubmed_entry["Id"] not in publications:
            publications[pubmed_entry["Id"]] = {"pubmed_id": int(pubmed_entry["Id"])}

            try:
                publications[pubmed_entry["Id"]]["title"] = str(pubmed_entry["Title"])
            except ValueError:
                pass

            try:
                publications[pubmed_entry["Id"]]["year"] = int(
                    pubmed_entry["PubDate"].split(" ")[0]
                )
            except ValueError:
                pass

            try:
                publications[pubmed_entry["Id"]]["author"] = (
                    [str(x) for x in pubmed_entry["AuthorList"][0:3]] + ["et. al"]
                    if len(pubmed_entry["AuthorList"]) > 3
                    else [str(x) for x in pubmed_entry["AuthorList"]]
                )
            except ValueError:
                pass

            try:
                publications[pubmed_entry["Id"]]["journal"] = str(
                    pubmed_entry["Source"]
                )
            except ValueError:
                pass

            try:
                publications[pubmed_entry["Id"]]["volume"] = int(pubmed_entry["Volume"])
            except ValueError:
                pass

            try:
                publications[pubmed_entry["Id"]]["issue"] = int(pubmed_entry["Issue"])
            except ValueError:
                pass

            try:
                publications[pubmed_entry["Id"]]["pages"] = str(pubmed_entry["Pages"])
            except ValueError:
                pass

            try:
                publications[pubmed_entry["Id"]]["doi"] = str(pubmed_entry["DOI"])
            except ValueError:
                pass

    return publications


def validate_part(
    elem, wi_object, warnings, pooled, organisms, settings, errors, pubmed_ids
):
    """
    This function is used to validate a part of the WI object. If an error or a
    warning is found than it is added to the description of the key whose value
    contains the error/warning. This function is called simultaneously for the
    parts 'project', 'experimental_setting' and 'technical_details'.
    :param elem: the key of the part that is being validated
    :param wi_object: the object that is validated
    :param warnings: a list containing all warnings
    :param pooled: a boolean that states if the samples were pooled
    :param organisms: a list of all contained organisms
    :param errors: a list containing all errors
    :return: elem: the key of the part that is being validated
             wi_object: the validated object containing error and warn messages
             pooled: a boolean stating if the samples are pooled
             organisms: a list of all contained organisms
             warnings: a list containing all found warnings
             errors: a list containing all found errors
    """
    error_desc = ""
    warning_desc = ""
    if isinstance(wi_object, dict):
        if "desc" in wi_object and "backup_desc" not in wi_object:
            wi_object["backup_desc"] = wi_object["desc"]

        if wi_object["list"]:
            if wi_object["position"].split(":")[-1] == "techniques":
                tech_settings = []
                for t_elem in wi_object["list_value"]:
                    for t_sub_elem in t_elem:
                        if t_sub_elem["position"].split(":")[-1] == "setting":
                            tech_settings.append(t_sub_elem["value"])
                warning, warn_text = validate_yaml.validate_techniques(
                    settings, tech_settings
                )
                if warning:
                    warnings.append(
                        {
                            "position": wi_object["position"],
                            "title": "Invalid Technique",
                            "message": warn_text,
                        }
                    )

                    warn_text = warn_text.replace("\n", "<br>")
                    warning_desc = (
                        f"{warning_desc}"
                        f'{"<br>" if warning_desc != "" else ""}'
                        f'<font color="orange">{warn_text}</font>'
                    )
                    wi_object["desc"] = (
                        f'{wi_object["backup_desc"]}'
                        f'{"<br>" if wi_object["backup_desc"] != "" else ""}'
                        f"{warning_desc}"
                    )
                else:
                    wi_object["desc"] = wi_object["backup_desc"]
            elif not any(
                [
                    isinstance(x, dict) or isinstance(x, list)
                    for x in wi_object["list_value"]
                ]
            ):
                error = False
                messages = []
                for sub_elem in wi_object["list_value"]:
                    valid, message = validate_yaml.validate_value(
                        sub_elem,
                        wi_object["input_type"],
                        wi_object["position"].split(":")[-1],
                    )
                    if not valid:
                        error = True
                        messages.append((sub_elem, message))
                        errors.append(
                            {
                                "position": wi_object["position"],
                                "title": f'Invalid {wi_object["displayName"]}',
                                "message": message,
                            }
                        )
                wi_object["error"] = error
                if error:
                    message = ", ".join([f"{msg[0]}: {msg[1]}" for msg in messages])
                    error_desc = (
                        f"{error_desc}"
                        f'{"<br>" if error_desc != "" else ""}'
                        f'<font color="red">{message}</font>'
                    )
                wi_object["desc"] = (
                    f'{wi_object["backup_desc"]}'
                    f'{"<br>" if wi_object["backup_desc"] != "" else ""}'
                    f'{error_desc}{"<br>" if error_desc != "" else ""}'
                    f"{warning_desc}"
                )
            else:
                (
                    elem,
                    wi_object["list_value"],
                    pooled,
                    organisms,
                    warnings,
                    errors,
                    pubmed_ids,
                ) = validate_part(
                    elem,
                    wi_object["list_value"],
                    warnings,
                    pooled,
                    organisms,
                    settings,
                    errors,
                    pubmed_ids,
                )
        else:
            if "input_fields" in wi_object:
                (
                    elem,
                    wi_object["input_fields"],
                    pooled,
                    organisms,
                    warnings,
                    errors,
                    pubmed_ids,
                ) = validate_part(
                    elem,
                    wi_object["input_fields"],
                    warnings,
                    pooled,
                    organisms,
                    settings,
                    errors,
                    pubmed_ids,
                )
            else:
                if (
                    wi_object["position"].split(":")[-1] == "pubmed_id"
                    and wi_object["value"] is not None
                ):
                    pubmed_ids.append(wi_object["value"])
                if wi_object["value"] is not None and wi_object["value"] != "":
                    if wi_object["input_type"] == "date":
                        value = wi_utils.date_to_str(wi_object["value"])
                    else:
                        value = wi_object["value"]
                    valid, message = validate_yaml.validate_value(
                        value,
                        wi_object["input_type"],
                        wi_object["position"].split(":")[-1],
                    )
                    wi_object["error"] = not valid
                    if not valid:
                        errors.append(
                            {
                                "position": wi_object["position"],
                                "title": f'Invalid {wi_object["displayName"]}',
                                "message": message,
                            }
                        )
                        error_desc = (
                            f"{error_desc}"
                            f'{"<br>" if error_desc != "" else ""}'
                            f'<font color="red">{message}</font>'
                        )
                    warning = False
                    warn_text = None
                    warn_title = ""
                    key = wi_object["position"].split(":")[-1]
                    if key == "pooled":
                        pooled = wi_object["value"]
                    elif key == "donor_count":
                        warning, warn_text = validate_yaml.validate_donor_count(
                            pooled, wi_object["value"]
                        )
                        warn_title = "Illogical donor count"
                    elif key == "organism":
                        organisms.append(wi_object["value"].split(" ")[0])
                    elif key == "reference_genome":
                        warning, warn_text = validate_yaml.validate_reference_genome(
                            organisms, wi_object["value"]
                        )
                        warn_title = "Illogical reference genome"
                    wi_object["warning"] = warning
                    if warning:
                        warnings.append(
                            {
                                "position": wi_object["position"],
                                "title": warn_title,
                                "message": warn_text,
                            }
                        )
                        warning_desc = (
                            f"{warning_desc}"
                            f'{"<br>" if warning_desc != "" else ""}'
                            f'<font color="orange">{warn_text}</font>'
                        )
                    wi_object["desc"] = (
                        f'{wi_object["backup_desc"]}'
                        f'{"<br>" if wi_object["backup_desc"] != "" else ""}'
                        f'{error_desc}{"<br>" if error_desc != "" else ""}'
                        f"{warning_desc}"
                    )
    elif isinstance(wi_object, list):
        for i in range(len(wi_object)):
            elem, wi_object[i], pooled, organisms, warnings, errors, pubmed_ids = (
                validate_part(
                    elem,
                    wi_object[i],
                    warnings,
                    pooled,
                    organisms,
                    settings,
                    errors,
                    pubmed_ids,
                )
            )
    return elem, wi_object, pooled, organisms, warnings, errors, pubmed_ids



import fred.src.utils as utils
import fred.src.web_interface.whitelist_parsing as whitelist_parsing
import fred.src.web_interface.yaml_to_wi_object as yto
import fred.src.web_interface.wi_utils as wi_utils
import copy
import re


def get_factors(organism, key_yaml, read_in_whitelists):
    """
    This function returns all experimental factors with a whitelist of their
    values
    :param key_yaml: the read in general structure
    :param organism: the organism that was selected by the user
    :return: factor_value: a dictionary containing factors and whitelists
    """

    factor_desc = list(utils.find_keys(key_yaml, "experimental_factors"))[0][
        "factor_desc"
    ]
    f_node = list(utils.find_keys(key_yaml, "factor"))[0]
    # initialize dictionary with all factors
    factor_list, whitelist_type, input_type, headers, whitelist_keys, double = (
        whitelist_parsing.parse_whitelist(
            "factor",
            f_node,
            {"organism": organism},
            whitelist_object=read_in_whitelists,
        )
    )

    factor_value = {"factor": factor_list, "desc": factor_desc}
    plain_factors = []

    for elem in factor_list:
        plain_factors += elem["whitelist"]
    # initialize empty dictionary to store values of each factor
    values = {}

    # iterate over factors
    for factor in plain_factors:

        # get attributes of factor from general structure
        node = list(utils.find_keys(key_yaml, factor))

        # factor was found in general structure
        if len(node) > 0:

            # call function 'get_factor_values' to get whitelist information
            (
                whitelist,
                whitelist_type,
                input_type,
                headers,
                w_keys,
                double,
                nested_infos,
            ) = get_factor_values(
                factor, node[0], {"organism": organism}, read_in_whitelists
            )

            # change single_autofill to multi_autofill because all factors can
            # have multiple values
            if input_type == "single_autofill":
                input_type = "multi_autofill"

            # save whitelist, input_type and whitelist_type for the values of
            # one factor
            values[factor] = {
                "whitelist": whitelist,
                "input_type": input_type,
                "whitelist_type": whitelist_type,
                "displayName": node[0]["display_name"],
                "desc": utils.print_desc(
                    (
                        node[0]["factor_desc"]
                        if "factor_desc" in node[0]
                        else node[0]["desc"]
                    ),
                    "html",
                ).replace("\n", "<br>"),
            }

            # add search_info if input is of type single- or multi-autofill
            if input_type == "multi_autofill":
                values[factor]["search_info"] = {
                    "organism": organism,
                    "key_name": factor,
                }

            if input_type == "restricted_short_text":
                # default
                values[factor]["restriction"] = {"regex": " ", "max_length": 10}

                if "special_case" in node and "restriction" in node["special_case"]:
                    if "regex" in node["special_case"]["restriction"]:
                        values[factor]["restriction"]["regex"] = node["special_case"][
                            "restriction"
                        ]["regex"]
                    if "max_length" in node["special_case"]["restriction"]:
                        values[factor]["restriction"]["max_length"] = node[
                            "special_case"
                        ]["restriction"]["max_length"]

            # add header and whitelist keys if they are defined
            if headers is not None:
                values[factor]["headers"] = headers
            if w_keys is not None:
                values[factor]["whitelist_keys"] = w_keys
            if len(double) > 0:
                values[factor]["double"] = double
            if nested_infos is not None:
                values[factor]["nested_infos"] = nested_infos

    # add the values to the dictionary
    factor_value["values"] = values

    return factor_value


def get_factor_values(key, node, filled_object, read_in_whitelists, nested_infos=None):
    """
    This function is used to get the whitelists of experimental factors
    including the whitelist type, input type, headers and whitelist keys
    :param key: the name of the experimental factor
    :param node: the part of the general structure containing information of
                 the factor
    :param filled_object: a dictionary storing the used organism and used for
                          parsing of whitelists of type 'depend' (dependent
                          on organism)
    :return:
    whitelist: a list or dictionary of available values for input (type depends
               on whitelist_type)
    whitelist_type: the type of the whitelist (e.g. plain, group, ...)
    input_type: the input type for the input field in the web interface (e.g.
                short_text, select, single_autofill, ...)
    headers: the headers of the whitelist (None if no headers are defined)
    w_keys: the keys of a whitelist of type group that was rewritten to type
            plain (None if the whitelist is not plain group)
    """

    # initialize headers, whitelist keys and whitelist type with None
    headers = None
    w_keys = None
    whitelist_type = None
    double = []

    # value is a dictionary and no special case
    if isinstance(node["value"], dict) and not (
        "special_case" in node
        and ("merge" in node["special_case"] or "value_unit" in node["special_case"])
    ):

        # initialize whitelist as empty list
        whitelist = []
        nested_infos = {}

        # iterate over the keys of the value
        for k in node["value"]:

            key_info = {}

            # initialize an empty dictionary to store the properties of
            # the key
            k_val = {}

            # call this function to get the whitelist information for the keys
            (
                k_val["whitelist"],
                k_val["whitelist_type"],
                k_val["input_type"],
                header,
                whitelist_keys,
                doubled,
                nested_infos,
            ) = get_factor_values(
                k,
                node["value"][k],
                filled_object,
                read_in_whitelists,
                nested_infos=nested_infos,
            )

            # add header and whitelist keys to dictionary if they are defined
            if header is not None:
                k_val["headers"] = header
                key_info["headers"] = header
            if whitelist_keys is not None:
                k_val["whitelist_keys"] = whitelist_keys
                key_info["whitelist_keys"] = whitelist_keys
            if len(doubled) > 0:
                k_val["double"] = doubled
                key_info["double"] = doubled

            # add key 'unit' if key is of special case value_unit
            if k_val["input_type"] == "value_unit":
                k_val["unit"] = None

            # add properties from the general structure
            k_val["displayName"] = node["value"][k]["display_name"]
            k_val["required"] = node["value"][k]["mandatory"]
            k_val["position"] = k
            k_val["value"] = []
            k_val["desc"] = node["value"][k]["desc"]

            # change single_autofill to multi_autofill because all factors can
            # have multiple values
            if k_val["input_type"] == "single_autofill":
                k_val["input_type"] = "multi_autofill"

            if k_val["input_type"] == "restricted_short_text":

                # default
                k_val["restriction"] = {"regex": " ", "max_length": 10}

                if (
                    "special_case" in node["value"][k]
                    and "restriction" in node["value"][k]["special_case"]
                ):
                    if "regex" in node["value"][k]["special_case"]["restriction"]:
                        k_val["restriction"]["regex"] = node["value"][k][
                            "special_case"
                        ]["restriction"]["regex"]
                    if "max_length" in node["value"][k]["special_case"]["restriction"]:
                        k_val["restriction"]["max_length"] = node["value"][k][
                            "special_case"
                        ]["restriction"]["max_length"]

            # add search info if the input is of type single- or multi-autofill
            if k_val["input_type"] == "multi_autofill":
                k_val["whitelist"] = None
                k_val["search_info"] = {
                    "organism": filled_object["organism"],
                    "key_name": k,
                }

            # add dictionary with properties of the key to the whitelist
            whitelist.append(k_val)
            if len(key_info) > 0:
                nested_infos[k] = key_info

        # set input type to nested
        input_type = "nested"
        if len(nested_infos) == 0:
            nested_infos = None

    # value is not a dictionary or special_case: merge or value_unit
    else:

        # read and parse whitelist
        whitelist, whitelist_type, input_type, headers, w_keys, double = (
            whitelist_parsing.parse_whitelist(
                key, node, filled_object, whitelist_object=read_in_whitelists
            )
        )

    # factor takes a list as value -> can occur multiple times in one condition
    if node["list"]:

        # values are single values (no dictionaries)
        if not isinstance(node["value"], dict):

            # change single_autofill to multi_autofill because all factors can
            # have multiple values
            if input_type == "single_autofill":
                input_type = "multi_autofill"

    return whitelist, whitelist_type, input_type, headers, w_keys, double, nested_infos


def get_conditions(factors, organism_name, key_yaml, read_in_whitelists):
    """
    This function creates all combinations of selected experimental factors and
    their values
    :param key_yaml: the read in general structure
    :param organism_name: the selected organism
    :param factors: multiple dictionaries containing the keys 'factor' and
    'values' with their respective values grouped in a list
    e.g. [{'factor': 'gender', 'values': ['male', 'female']},
          {'factor: 'life_stage', 'values': ['child', 'adult']}]
    :return: a dictionary containing the following keys:
             conditions: a condition object containing a list of all conditions
                         saved as dictionaries
             whitelist_object: a dictionary containing all keys that store a
                               whitelist and their whitelist
             organism: the organism that was chosen by the user
    """

    # initialize an empty dictionary to store the values as they are displayed
    # on the website -> used if there are headers or whitelist keys
    # e.g. value 'GFP' is displayed as 'GFP (other)' on the website since the
    # value is stored in a grouped whitelist that was refactored to plain
    real_val = {}

    # iterate over factors
    for i in range(len(factors)):

        # extract the properties of the factor from the general structure
        factor_infos = list(utils.find_keys(key_yaml, factors[i]["factor"]))

        # sanity check -> factor was found in general structure and
        # list of values contains only one element
        if len(factor_infos) > 0 and len(factors[i]["values"]) == 1:

            # set val to the values specified for the factor
            val = factors[i]["values"][0]

            # value is a dictionary -> e.g. disease
            if isinstance(val, dict):

                # remove keys with value None or empty lists and dictionaries
                val = {
                    k: v
                    for k, v in val.items()
                    if not (type(v) in [list, dict] and len(v) == 0) and v is not None
                }

                # add an ident key to the value
                # -> defined in general structure or None
                val["ident_key"] = (
                    factor_infos[0]["special_case"]["group"]
                    if "special_case" in factor_infos[0]
                    and "group" in factor_infos[0]["special_case"]
                    else None
                )

                val["control"] = (
                    factor_infos[0]["special_case"]["control"]
                    if "special_case" in factor_infos[0]
                    and "control" in factor_infos[0]["special_case"]
                    else None
                )

                if "nested_infos" in factors[i]:
                    for val_key in factors[i]["nested_infos"]:
                        if val_key in val:
                            for v in range(len(val[val_key])):
                                full_value = copy.deepcopy(val[val_key][v])
                                whitelist_key = None
                                if (
                                    "whitelist_keys"
                                    in factors[i]["nested_infos"][val_key]
                                ):
                                    for w_key in factors[i]["nested_infos"][val_key][
                                        "whitelist_keys"
                                    ]:
                                        if val[val_key][v].endswith(f"({w_key})"):
                                            val[val_key][v] = (
                                                val[val_key][v]
                                                .rstrip(f"({w_key})")
                                                .strip()
                                            )
                                            whitelist_key = w_key
                                            break
                                if "headers" in factors[i]["nested_infos"][val_key]:
                                    if isinstance(
                                        factors[i]["nested_infos"][val_key]["headers"],
                                        dict,
                                    ):
                                        if (
                                            whitelist_key is not None
                                            and whitelist_key
                                            in factors[i]["nested_infos"][val_key][
                                                "headers"
                                            ]
                                        ):
                                            str_value = wi_utils.parse_headers(
                                                factors[i]["nested_infos"][val_key][
                                                    "headers"
                                                ][whitelist_key],
                                                val[val_key][v],
                                                mode="str",
                                            )
                                            val[val_key][v] = (
                                                f'{"{"}' f'{str_value}{"}"}'
                                            )
                                    else:
                                        str_value = wi_utils.parse_headers(
                                            factors[i]["nested_infos"][val_key][
                                                "headers"
                                            ],
                                            val[val_key][v],
                                            mode="str",
                                        )
                                        val[val_key][v] = f'{"{"}' f'{str_value}{"}"}'
                                real_val[val[val_key][v]] = full_value

                # generate combinations of the values of the dictionary for the
                # conditions and overwrite the values with them
                factors[i]["values"] = utils.get_combis(
                    val,
                    factors[i]["factor"],
                    {"organism": organism_name.split(" ")[0]},
                    key_yaml,
                    read_in_whitelists=read_in_whitelists,
                )

        elif factor_infos[0]["list"]:

            new_values = []
            for j in range(len(factors[i]["values"])):
                single_val = copy.deepcopy(factors[i]["values"][j])
                single_val_key = None
                if "whitelist_keys" in factors[i]:
                    for w_key in factors[i]["whitelist_keys"]:
                        if single_val.endswith(f"({w_key})"):
                            single_val = single_val.rstrip(f"({w_key})").strip()
                            single_val_key = w_key
                            break
                if "headers" in factors[i]:
                    if isinstance(factors[i]["headers"], dict):
                        if single_val_key is not None:
                            if single_val_key in factors[i]["headers"]:
                                single_val = wi_utils.parse_headers(
                                    factors[i]["headers"][single_val_key],
                                    single_val,
                                    mode="dict",
                                )
                    else:
                        single_val = wi_utils.parse_headers(
                            factors[i]["headers"], single_val, mode="dict"
                        )
                if isinstance(single_val, dict):
                    val = "|".join([f'{key}:"{single_val[key]}"' for key in single_val])
                    val = f'{factors[i]["factor"]}:{"{"}{val}{"}"}'
                    real_val[val] = factors[i]["values"][j]
                new_values.append(single_val)
            factors[i]["values"] = utils.get_combis(
                new_values,
                factors[i]["factor"],
                {"organism": organism_name.split(" ")[0]},
                key_yaml,
                read_in_whitelists=read_in_whitelists,
            )

        # iterate over all values
        for j in range(len(factors[i]["values"])):

            # factor contains whitelist keys
            if "whitelist_keys" in factors[i]:

                # save the original value as full_value
                full_value = copy.deepcopy(factors[i]["values"][j])

                # save the headers if they are specified for the factor else
                # None
                headers = factors[i]["headers"] if "headers" in factors[i] else None

                # rewrite the values by removing the whitelist keys and split
                # them according to the headers if headers were defined
                factors[i]["values"][j] = wi_utils.parse_whitelist_keys(
                    factors[i]["whitelist_keys"],
                    factors[i]["values"][j],
                    headers,
                    mode="str",
                )

                # headers were defined and the whitelist key of the value is
                # defined within the headers
                if (
                    headers is not None
                    and full_value.split("(")[-1].replace(")", "") in headers
                ):
                    # rewrite the value to '<factor>:{<values>}'
                    factors[i]["values"][j] = (
                        f'{factors[i]["factor"]}:{"{"}'
                        f'{factors[i]["values"][j]}{"}"}'
                    )

                # save the original value in real_val with the new value as key
                if factors[i]["values"][j] not in real_val:
                    real_val[factors[i]["values"][j]] = full_value

            # factor contains headers but not whitelist keys
            elif "headers" in factors[i]:

                # save the original value
                full_value = copy.deepcopy(factors[i]["values"][j])

                # split the value according to the header and save them as a
                # string
                str_value = wi_utils.parse_headers(
                    factors[i]["headers"], factors[i]["values"][j], mode="str"
                )

                # # rewrite the value to '<factor>:{<values>}'
                factors[i]["values"][j] = (
                    f'{factors[i]["factor"]}:{"{"}' f'{str_value}{"}"}'
                )

                # save the original value in real_val with the new value as key
                if factors[i]["values"][j] not in real_val:
                    real_val[factors[i]["values"][j]] = full_value

    # generate the conditions
    conditions = utils.get_condition_combinations(factors)

    # extract the properties of 'sample' from the general structure
    sample = list(utils.find_keys(key_yaml, "samples"))

    # initialize a condition- and a whitelist object
    condition_object = []
    whitelist_object = {}

    # sanity check -> sample was found in the general structure
    if len(sample) > 0:

        # create an emtpy wi_object and a whitelist object from the sample
        # structure
        sample, whitelist_object = yto.parse_empty(
            sample[0],
            "experimental_setting:conditions:biological_replicates:samples",
            key_yaml,
            {"organism": organism_name},
            read_in_whitelists,
            get_whitelist_object=True,
        )

        # overwrite sample with its input fields
        sample_desc = sample["desc"]
        sample = sample["input_fields"]

        # iterate over conditions
        for cond in conditions:

            # generate a sample name from the condition
            sample_name = utils.get_short_name(
                cond,
                {"organism": organism_name},
                key_yaml,
                read_in_whitelists=read_in_whitelists,
            )

            # split the condition into key-value pairs
            split_condition = utils.split_cond(cond)
            search_condition = []

            # TODO: remove list (can there be a list?) or own function for items
            for elem in split_condition:
                if isinstance(elem[1], dict):
                    val = "|".join([f'{k}:"{elem[1][k]}"' for k in elem[1]])
                    val = f'{elem[0]}:{"{"}{val}{"}"}'
                    if val in real_val:
                        search_condition.append(f'{elem[0]}:"{real_val[val]}"'.lower())
                    else:
                        for key in elem[1]:
                            if isinstance(elem[1][key], dict):
                                val = "|".join(
                                    [f'{k}:"{elem[1][key][k]}"' for k in elem[1][key]]
                                )
                                val = f'{key}:{"{"}{val}{"}"}'
                                if val in real_val:
                                    search_condition.append(
                                        f'{elem[0]}:"{real_val[val]}"'.lower()
                                    )
                                else:
                                    # TODO: real/val or own function to get values from deeper dictionaries
                                    for sub_key in elem[1][key]:
                                        search_condition.append(
                                            f'{elem[0]}:"{elem[1][key][sub_key]}"'.lower()
                                        )
                            else:
                                search_condition.append(
                                    f'{elem[0]}:"{elem[1][key]}"'.lower()
                                )
                elif isinstance(elem[1], list):
                    for item in elem[1]:
                        search_condition.append(f'{elem[0]}:"{item}"'.lower())
                else:
                    if elem[1] in real_val:
                        search_condition.append(
                            f'{elem[0]}:"{real_val[elem[1]]}"'.lower()
                        )
                    else:
                        search_condition.append(f'{elem[0]}:"{elem[1]}"'.lower())
            # call functions to fill the samples with the values from the
            # condition
            filled_sample = get_samples(
                split_condition,
                copy.deepcopy(sample),
                real_val,
                key_yaml,
                sample_name,
                organism_name,
                read_in_whitelists,
            )

            # rewrite condition into title and string to readd deleted
            # conditions
            title, readd = get_condition_title(split_condition)

            # save the condition as a dictionary with the filled sample as
            # input fields
            d = {
                "correct_value": cond,
                "title": title,
                "search": search_condition,
                "readd": readd,
                "position": "experimental_setting:condition",
                "list": True,
                "mandatory": True,
                "list_value": [],
                "input_disabled": False,
                "desc": sample_desc,
                "input_fields": copy.deepcopy(filled_sample),
            }

            # add the dictionary for the condition to the condition object
            condition_object.append(d)

    return {
        "conditions": condition_object,
        "whitelist_object": whitelist_object,
        "organism": organism_name,
    }


def get_condition_title(split_condition):
    """
    This function rewrites a condition into a table to be displayed as a title
    on the website. It also creates a string that is shown when removed
    conditions have to be re-added
    :param split_condition: a list containing the factors and values the
                            condition consists of
    :return: html: the title as a table in html
             readd: the string shown when re-adding deleted conditions
    """

    # initialize the html string with the table header and style
    html = '<table class="table_style_condition_title"><tbody>'

    # initialize readd as an empty string
    readd = ""

    # iterate over the factors and values
    for i in range(len(split_condition)):

        # start the row of the table with the style depending on the index of
        # the element that is written into the row (the style creates a
        # horizontal line below the row and includes padding)
        if len(split_condition) > 1 and i < len(split_condition) - 1:
            html += '<tr class="tr_style_condition_title">'
        else:
            html += "<tr>"

        # add the factor to readd
        readd += f"{split_condition[i][0]}:\n"

        # add the factor to the table as a data cell with the style of a value
        # (padding)
        html += (
            f'<td class="td_style_condition_title_value">'
            f"{split_condition[i][0]}:</td>"
        )

        vals, readd = get_title_value(split_condition[i][1], readd)

        # add the string with the values to the table as a data cell
        html += f'<td class="td_style_condition_title_value">{vals}' f"</td></tr>"

        # if the condition consists of multiple factors than write a '---'
        # between them in readd
        if len(split_condition) > 1 and i < len(split_condition) - 1:
            readd += "---\n"

    # close the html table
    html += f"</tbody></table>"

    return html, readd


def get_title_value(cond_value, readd):
    # value is a dictionary
    if isinstance(cond_value, dict):

        # initialize an empty string to save the values to
        vals = ""

        # iterate over the keys
        for key in cond_value:

            value, readd = get_title_value(cond_value[key], readd)

            # add the value to the string with a <br> if it is not in the
            # last row
            if key != list(cond_value.keys())[-1]:
                vals += f"{value}<br>"
            else:
                vals += f"{value}"

    # value is not a dictionary
    else:

        # replace more than 3 zeros with 3 dots to shorten the value
        vals = f'"{re.sub("0000(0)*", "...", cond_value)}"'

        # add the value to readd
        readd += f"{vals}\n"

    return vals, readd


def get_samples(
    split_condition,
    sample,
    real_val,
    key_yaml,
    sample_name,
    organism_name,
    read_in_whitelists,
    is_factor=True,
):
    """
    This function created a pre-filled object with the structure of the samples
    to be displayed in the web interface
    :param is_factor: a boolean defining if the input field is an experimental
                      factor
    :param organism_name: the name of the selected organism
    :param sample_name: the identifier of the sample build from the condition
    :param key_yaml: the read in general structure
    :param real_val: a dictionary containing the values containing headers and
                     whitelist keys as they should be displayed on the website
    :param split_condition: the condition the sample is created for
    :param sample: the empty structure of the sample
    :return: sample: the pre-filled sample
    """
    # save all factors in a list
    factors = [cond[0] for cond in split_condition]

    # iterate over samples
    for i in range(len(sample)):

        # input field: sample_name
        if sample[i]["position"].endswith("samples:sample_name"):

            # TODO: improve display
            # add whitespaces to sample name to enable line breaks on the
            # website
            sample[i]["value"] = (
                sample_name.replace(":", ": ")
                .replace("|", "| ")
                .replace("#", "# ")
                .replace("-", " - ")
                .replace("+", " + ")
            )

            # save the unchanged sample name as 'correct_value'
            sample[i]["correct_value"] = sample_name.split("_")[0]

        # elif sample[i]['position'].endswith('samples:sequencer_name'):
        #    sample[i]['value'] = sequencer_name
        #    sample[i]['input_disabled'] = True

        # input field is a factor
        elif sample[i]["position"].split(":")[-1] in factors:

            # iterate over factors in condition
            for c in split_condition:

                # input field of current factor
                if sample[i]["position"].split(":")[-1] == c[0]:
                    # extract properties of the factor from the general
                    # structure
                    info = list(utils.find_keys(key_yaml, c[0]))

                    # sanity check -> factor was found in general structure
                    if len(info) > 0:

                        filled_value = ""
                        # special case value_unit
                        if (
                            "special_case" in info[0]
                            and "value_unit" in info[0]["special_case"]
                        ):

                            if isinstance(c[1], dict):
                                value_unit = c[1]
                            else:
                                # split the value into value and unit
                                value_unit = utils.split_value_unit(c[1])

                            # save the value and unit in the sample
                            sample[i]["value"] = value_unit["value"]
                            sample[i]["value_unit"] = value_unit["unit"]
                            if is_factor:
                                sample[i]["input_disabled"] = True

                        # value is a dictionary
                        elif isinstance(c[1], dict):

                            # rewrite the value into a string
                            val = "|".join([f'{key}:"{c[1][key]}"' for key in c[1]])
                            val = f'{c[0]}:{"{"}{val}{"}"}'
                            # save the value from the dictionary real_val if
                            # real_val contains the current value as key

                            if val in real_val:
                                filled_value = real_val[val]

                            # value is not in real_val
                            else:

                                # TODO: rework with real_val

                                if "input_fields" in sample[i]:
                                    # call this function on the keys of the
                                    # value in order to fill them
                                    filled_value = get_samples(
                                        [
                                            (x, c[1][x])
                                            for x in c[1]  # if not (
                                            # c[1] == 'technical_replicates'
                                            # and x == 'sample_name')
                                        ],
                                        copy.deepcopy(sample[i]["input_fields"]),
                                        info,
                                        key_yaml,
                                        sample_name,
                                        organism_name,
                                        read_in_whitelists,
                                        is_factor=is_factor,
                                    )
                                # TODO: als real_val? automate
                                elif c[0] == "enrichment_type":
                                    connected_value = []
                                    for key in c[1]:
                                        connected_value.append(c[1][key])
                                    connected_value = " ".join(connected_value)
                                    filled_value = f"{connected_value} (proteins)"
                                    sample[i]["headers"] = {
                                        "proteins": "gene_name ensembl_id"
                                    }
                                elif "headers" in sample[i]:
                                    headers = [x for x in c[1]]
                                    w_key = None
                                    if isinstance(sample[i]["headers"], dict):

                                        for k in sample[i]["headers"]:
                                            if sorted(
                                                sample[i]["headers"][k].split(" ")
                                            ) == sorted(headers):
                                                w_key = k
                                                break
                                    else:
                                        if sorted(headers) != sorted(
                                            sample[i]["headers"].split(" ")
                                        ):
                                            headers = None

                                    if headers is not None:
                                        filled_value = ""
                                        for header in headers:
                                            filled_value = (
                                                filled_value + " " + c[1][header]
                                            )
                                        filled_value = filled_value.lstrip(" ").rstrip(
                                            " "
                                        )

                                    if (
                                        filled_value is not None
                                        and w_key is not None
                                        and sample[i]["whitelist_type"] == "plain_group"
                                    ):
                                        filled_value = f"{filled_value} " f"({w_key})"
                            # save the filled value in 'list_value' if the
                            # input field takes a list
                            if sample[i]["list"]:
                                sample[i]["list_value"].append(filled_value)

                            # save the filled value in 'input_fields' if the
                            # input field takes a dictionary
                            elif "input_fields" in sample[i]:
                                sample[i]["input_fields"] = filled_value

                            # save the filled value in 'value'
                            else:
                                sample[i]["value"] = filled_value

                        # value is not a dictionary
                        else:
                            # save the value from the dictionary real_val if it
                            # contains the current value
                            if (
                                c[0] == "enrichment_type"
                                and "headers" in sample[i]
                                and "proteins" in sample[i]["headers"]
                            ):
                                if not c[1].endswith("(proteins)"):
                                    filled_value = f"{c[1]} (proteins)"
                                else:
                                    filled_value = c[1]

                            elif c[1] in real_val:
                                filled_value = real_val[c[1]]

                            # save the current value
                            else:

                                if (
                                    sample[i]["whitelist_type"] == "plain_group"
                                    and "whitelist_keys" in sample[i]
                                ):
                                    w = utils.get_whitelist(
                                        sample[i]["position"].split(":")[-1],
                                        {"organism": organism_name},
                                        whitelist_object=read_in_whitelists,
                                    )

                                    for key in sample[i]["whitelist_keys"]:
                                        filled_value = c[1]
                                        if f"{c[1]} ({key})" in w["whitelist"]:
                                            filled_value = f"{c[1]} ({key})"
                                            break
                                else:
                                    if (
                                        "input_type" in info[0]
                                        and info[0]["input_type"] == "number"
                                    ):
                                        filled_value = int(c[1])
                                    else:
                                        filled_value = c[1]
                            # save the filled value in 'list_value' if the
                            # input field takes a list
                            if sample[i]["list"]:
                                sample[i]["list_value"].append(filled_value)

                            # save the filled value in 'value'
                            else:
                                sample[i]["value"] = filled_value

                        # input field is of type single_autofill
                        if "input_type" in sample[i] and sample[i]["input_type"] in [
                            "single_autofill",
                            "restricted_short_text",
                        ]:
                            # initialize the key 'list_value' and move the
                            # value under the key 'value' to the key
                            # 'list_value'
                            sample[i]["list_value"] = (
                                []
                                if sample[i]["value"] is None
                                else [sample[i]["value"]]
                            )
                            sample[i]["value"] = None

                        if is_factor:

                            if "list" in info[0] and info[0]["list"]:
                                sample[i]["delete_disabled"] = True
                                sample[i]["fixed_length"] = len(sample[i]["list_value"])
                            else:
                                # disable the input for the filled input field
                                sample[i]["input_disabled"] = True

                        if "input_type" in info[0] and info[0]["input_type"] == "bool":
                            if sample[i]["value"]:
                                sample[i]["value"] = "True"
                            else:
                                sample[i]["value"] = "False"

    return sample



from dash import Dash, html, dcc, callback, Output, Input
import plotly.express as px
import plotly.graph_objects as go
import plotly as plt
import pandas as pd
import argparse
import pathlib
from fred.src import utils
from fred.src.heatmap.dash_utils import *
import numpy as np
from fred.src.heatmap.dash_utils import get_data
import os
import base64
import io
from PIL import Image

parser = argparse.ArgumentParser("Metadata Heatmap")
parser.add_argument(
    "-p",
    "--path",
    type=pathlib.Path,
    required=True,
    help="The path of the metadata file",
)

args = parser.parse_args()
input_file = utils.read_in_yaml(args.path)
key_yaml = utils.read_in_yaml("keys.yaml")

settings, experimental_factors, organisms, max_vals, options_pretty, annotated_dict = (
    get_data(input_file, key_yaml)
)


app = Dash()

# Requires Dash 2.17.0 or later
app.layout = [
    dcc.Tabs(
        id="tabs-example-graph",
        value=f"tab-{list(settings.keys())[0]}",
        children=[dcc.Tab(label=f"Setting {x}", value=f"tab-{x}") for x in settings],
    ),
    dcc.Checklist(["Show Empty"], [], id="checklist-selection"),
    dcc.Graph(id="graph-content"),
]


@callback(
    Output("graph-content", "figure"),
    Input("tabs-example-graph", "value"),
    Input("checklist-selection", "value"),
)
def update_graph(value, empty):
    value = value.replace("tab-", "")
    options = [
        key.replace("_num", "") for key in settings[value] if key.endswith("_num")
    ]
    sorter = experimental_factors[value] + [
        o for o in options if o not in experimental_factors[value]
    ]

    if not empty:
        df_empty = settings[value].dropna(axis=1, how="all")
        sorter = [x for x in sorter if x in df_empty.columns]
    df = [settings[value][f"{key}_num"] for key in sorter]

    annotated = [
        annotated_dict[value][key] for key in sorter if key in annotated_dict[value]
    ]
    label_text = []
    for key in sorter:
        if key in annotated_dict[value] and key in experimental_factors[value]:
            label_text.append(annotated_dict[value][key])
        else:
            label_text.append("")

    option_text = []

    for option in sorter:
        if option in experimental_factors[value]:
            option_text.append(color("red", f"<b>{options_pretty[option]}</b>"))
        else:
            option_text.append(color("black", options_pretty[option]))

    plotly_colors = [x for x in px.colors.qualitative.G10 if x != "#DC3912"]
    print(plotly_colors)
    colors = [[0, "white"]]
    for i in range(1, max_vals[value] + 1):
        colors.append([i * 1 / (max_vals[value]), plotly_colors[i - 1]])
        if i * 1 / (max_vals[value]) != 1:
            colors.append([((i * 1) / (max_vals[value])) + 0.001, "white"])

    heatmap = [
        go.Heatmap(
            z=df,
            zmin=0,
            zmax=max_vals[value],
            x=[settings[value]["condition_index"], settings[value]["sample_index"]],
            y=sorter,
            showscale=False,
            customdata=annotated,
            text=label_text,
            texttemplate="%{text}",
            hovertemplate="%{customdata}",
            hoverongaps=False,
            colorscale=colors,
        ),
    ]

    condition_labels = {}
    for i in range(len(settings[value]["condition_name"])):
        if settings[value]["condition_index"][i] not in condition_labels:
            cond = settings[value]["condition_name"][i]
            splitted = utils.split_cond(cond)
            cond_dict = {}
            for elem in splitted:
                if isinstance(elem[1], dict):
                    vals = [elem[1][k] for k in elem[1]]
                else:
                    vals = [elem[1]]
                if elem[0] in cond_dict:
                    cond_dict[elem[0]] += vals
                else:
                    cond_dict[elem[0]] = vals
            condition_labels[settings[value]["condition_index"][i]] = cond_dict

    data_input = heatmap

    my_cell_width = 150
    top_margin = 100
    bottom_margin = 0
    left_margin = 200
    right_margin = 200
    my_height = 50 * len(sorter)
    my_width = my_cell_width * len(settings[value]["sample_index"])

    organism_path = os.path.join(
        os.path.dirname(__file__), "images", f"{organisms[value]}.png"
    )
    images = None

    if os.path.isfile(organism_path):
        plotly_logo = base64.b64encode(open(organism_path, "rb").read())
        imgdata = base64.b64decode(plotly_logo)
        im = Image.open(io.BytesIO(imgdata))
        im_width, im_height = im.size
        y_side = top_margin
        my_ysize = y_side / my_height
        x_side = y_side * im_width / im_height
        my_xsize = x_side / my_width

        left_margin = max(200, x_side)

        images = [
            dict(
                source="data:image/png;base64,{}".format(plotly_logo.decode()),
                xref="paper",
                yref="paper",
                x=0,
                y=1,
                sizey=my_ysize,
                sizex=my_xsize,
                xanchor="right",
                yanchor="bottom",
            )
        ]

    layout = go.Layout(
        images=images,
        height=my_height + top_margin + bottom_margin,
        width=my_width + left_margin + right_margin,
        margin=dict(l=left_margin, r=right_margin, t=top_margin, b=bottom_margin),
        autosize=False,
        title=dict(
            text=f"<b>Setting {value}</b>",
            font=dict(size=15),
            automargin=False,
            yref="container",
            x=(left_margin + (0.1 * 150)) / (my_width + left_margin + right_margin),
            y=1,
            xanchor="left",
            yanchor="top",
            subtitle=dict(
                text=f"Organism: {organisms[value]}",
                font=dict(size=14, lineposition="under"),
            ),
        ),
        plot_bgcolor="rgba(0,0,0,0)",
        paper_bgcolor="rgba(0,0,0,0)",
        xaxis=dict(side="top", showline=True, gridcolor="lightgrey", automargin=False),
        yaxis=dict(
            tickmode="array",
            ticktext=option_text,
            tickvals=sorter,
            autorange="reversed",
            tickson="boundaries",
            showline=True,
            gridcolor="lightgrey",
            automargin=False,
        ),
        legend=dict(
            # title=f"Organism: {organisms[value]}",
            orientation="h",
            x=0,
            y=0,
        ),
    )
    fig = go.Figure(data=data_input, layout=layout)

    for i in range(len(sorter)):
        fig.add_hline(i, line_dash="dash", line_color="lightgrey", layer="below")
        fig.add_hline(i - 0.5, line_width=0.5)
    fig.add_hline(len(sorter) - 0.5, line_width=0.5)

    for i in range(len(settings[value]["sample_index"]) + 1):
        fig.add_vline(i - 0.5, line_width=0.5)

    fig.add_hrect(
        y0=-0.5,
        y1=len(experimental_factors[value]) - 0.5,
        line=dict(color="red", width=5),
        layer="above",
    )

    return fig


def color(color, text):
    return f"<span style='color:{str(color)}'> {str(text)} </span>"


if __name__ == "__main__":

    app.run(debug=True)
